import sys
from pathlib import Path
from typing import Optional

import numpy as np
import pytorch_lightning as pl
import torch
import torchaudio
from everyvoice.config.type_definitions import TargetTrainingTextRepresentationLevel
from everyvoice.model.feature_prediction.config import FeaturePredictionConfig
from everyvoice.model.vocoder.HiFiGAN_iSTFT_lightning.hfgl.utils import (
    load_hifigan_from_checkpoint,
    synthesize_data,
)
from everyvoice.text.features import N_PHONOLOGICAL_FEATURES
from everyvoice.text.lookups import LookupTable
from everyvoice.text.text_processor import TextProcessor
from everyvoice.utils.heavy import expand
from loguru import logger
from torch import nn
from torchaudio.models import Conformer

from .config import FastSpeech2Config
from .layers import PositionalEmbedding, PostNet
from .loss import FastSpeech2Loss
from .noam import NoamLR
from .type_definitions_heavy import InferenceControl, Stats
from .utils.heavy import mask_from_lens, plot_attn_maps, plot_mel
from .variance_adaptor import VarianceAdaptor

DEFAULT_LANG2ID: LookupTable = {}
DEFAULT_SPEAKER2ID: LookupTable = {}


class FastSpeech2(pl.LightningModule):
    def __init__(
        self,
        config: dict | FastSpeech2Config,
        stats: Optional[dict | Stats] = None,
        lang2id: LookupTable = DEFAULT_LANG2ID,
        speaker2id: LookupTable = DEFAULT_SPEAKER2ID,
    ):
        """ """
        super().__init__()
        if not isinstance(config, FastSpeech2Config):
            config = FastSpeech2Config(**config)
        if stats is not None and not isinstance(stats, Stats):
            stats = Stats(**stats)
        self.config = config
        self.batch_size = config.training.batch_size
        self.text_processor = TextProcessor(config.text)
        # TODO Should we fallback to a default lang2id/speaker2id if we are loading an old model?
        self.lang2id = lang2id
        self.speaker2id = speaker2id
        self.stats = stats
        self.save_hyperparameters(ignore=[])
        self.loss = FastSpeech2Loss(config=config)
        self.text_input_layer: nn.Linear | nn.Embedding
        if (
            self.config.model.target_text_representation_level
            == TargetTrainingTextRepresentationLevel.phonological_features
        ):
            self.text_input_layer = nn.Linear(
                N_PHONOLOGICAL_FEATURES,
                self.config.model.encoder.input_dim,
                bias=False,
            )
        else:
            self.text_input_layer = nn.Embedding(
                len(self.text_processor.symbols),
                self.config.model.encoder.input_dim,
                padding_idx=self.text_processor.encode_text(
                    self.text_processor._pad_symbol
                )[0],
            )
        self.position_embedding = PositionalEmbedding(
            self.config.model.encoder.input_dim
        )

        self.encoder = Conformer(
            input_dim=self.config.model.encoder.input_dim,
            num_heads=self.config.model.encoder.heads,
            ffn_dim=self.config.model.encoder.feedforward_dim,
            num_layers=self.config.model.encoder.layers,
            depthwise_conv_kernel_size=self.config.model.encoder.conv_kernel_size,
            dropout=self.config.model.encoder.dropout,
        )
        if self.stats is None:
            logger.error(
                """Your model doesn't have a value for self.stats either because the file is missing or the checkpoint didn't save them.
                              We cannot initialize the variance adaptors without variance predictor statistics."""
            )
            self.variance_adaptor = None
        else:
            self.variance_adaptor = VarianceAdaptor(self.config, self.stats)

        self.decoder = Conformer(
            input_dim=self.config.model.decoder.input_dim,
            num_heads=self.config.model.decoder.heads,
            ffn_dim=self.config.model.decoder.feedforward_dim,
            num_layers=self.config.model.decoder.layers,
            depthwise_conv_kernel_size=self.config.model.decoder.conv_kernel_size,
            dropout=self.config.model.decoder.dropout,
        )

        self.mel_linear = nn.Linear(
            self.config.model.decoder.input_dim, self.config.preprocessing.audio.n_mels
        )  # TODO: replace with option for linear spec or complex
        if self.config.model.use_postnet:
            self.postnet = PostNet(
                n_mel_channels=self.config.preprocessing.audio.n_mels
            )  # TODO: allow for postnet parameterization in config
            self.output_key = "postnet_output"
        else:
            self.output_key = "output"
        self.speaker_embedding = None
        if self.config.model.multispeaker:
            if len(self.speaker2id) == 0:
                logger.error(
                    "Your model is multispeaker but speaker2id LookupTable is empty"
                )
                sys.exit(1)
            self.speaker_embedding = nn.Embedding(
                len(self.speaker2id),
                self.config.model.encoder.input_dim,
            )  # TODO: replace with d_vector multispeaker embedding
        self.language_embedding = None
        if self.config.model.multilingual:
            if len(self.lang2id) == 0:
                logger.error(
                    "Your model is multilingual but language2id LookupTable is empty"
                )
                sys.exit(1)
            self.language_embedding = nn.Embedding(
                len(self.lang2id), self.config.model.encoder.input_dim
            )

    def forward(self, batch, control=InferenceControl(), inference=False):
        # Add duration control. TODO: currently we just take the first item of the batch
        # and update the control accordingly, but this is hacky and should be fixed
        if "duration_control" in batch and batch["duration_control"][0]:
            control.duration = batch["duration_control"][0]

        # Determine whether we're teacher forcing or not
        # To do so, we need to be in inference mode and
        # the data loader should have loaded some Mel lengths
        # to force the output to match.
        if inference and batch["mel_lens"] is not None:
            teacher_forcing = True
        else:
            teacher_forcing = False

        # For model diagram see https://github.com/ming024/FastSpeech2/blob/master/img/model.png
        src_lens = batch["src_lens"]
        max_src_len = batch["max_src_len"]
        mel_lens = batch["mel_lens"]
        max_mel_len = batch["max_mel_len"]
        if (
            self.config.model.target_text_representation_level
            == TargetTrainingTextRepresentationLevel.phonological_features
        ):
            text_inputs = batch["pfs"]
        else:
            text_inputs = batch["text"]
        src_mask = mask_from_lens(src_lens, max_src_len)
        speaker_ids = batch["speaker_id"]
        language_ids = batch["language_id"]
        # Text Embedding
        inputs = self.text_input_layer(text_inputs)

        # Positional Embedding
        enc_pos_seq = torch.arange(
            max_src_len, device=inputs.device, requires_grad=False
        ).to(inputs.dtype)

        enc_pos_emb = self.position_embedding(enc_pos_seq) * src_mask.unsqueeze(2)

        # Encoder
        x, _ = self.encoder(inputs + enc_pos_emb, src_lens)  # expects B, T, K
        # Speaker Embedding
        if self.config.model.multispeaker and self.speaker_embedding:
            speaker_emb = self.speaker_embedding(speaker_ids)
            x = x + speaker_emb.unsqueeze(1)

        # Language Embedding
        if self.config.model.multilingual and self.language_embedding:
            lang_emb = self.language_embedding(language_ids)
            x = x + lang_emb.unsqueeze(1)

        # VarianceAdaptor out
        variance_adaptor_out = self.variance_adaptor(
            inputs,
            x,
            batch,
            src_mask,
            control,
            inference=inference,
            teacher_forcing=teacher_forcing,
        )
        # Create inference Mel lens
        if inference and not teacher_forcing:
            mel_lens = torch.IntTensor(
                [x.nonzero().size(0) for x in variance_adaptor_out["target_mask"]]
            ).to(text_inputs.device)
            max_mel_len = max(mel_lens)

        # Positional Embedding
        dec_pos_seq = torch.arange(max_mel_len, device=text_inputs.device).to(
            torch.float32 if inference else batch["mel"].dtype
        )
        dec_pos_emb = self.position_embedding(dec_pos_seq) * variance_adaptor_out[
            "target_mask"
        ].unsqueeze(2)

        # Decoder
        x, _ = self.decoder(variance_adaptor_out["output"] + dec_pos_emb, mel_lens)

        # Mel Linear
        output = self.mel_linear(x)

        # Postnet
        postnet_output = None
        if self.config.model.use_postnet:
            postnet_output = output + self.postnet(output)

        return {
            "output": output,
            "postnet_output": postnet_output,
            "src_mask": src_mask,
            "src_lens": src_lens,
            "tgt_mask": variance_adaptor_out["target_mask"],
            "tgt_lens": mel_lens,
            "attn_logprob": variance_adaptor_out["attn_logprob"],
            "attn_soft": variance_adaptor_out["attn_soft"],
            "attn_hard": variance_adaptor_out["attn_hard"],
            "duration_prediction": variance_adaptor_out["duration_prediction"],
            "duration_target": variance_adaptor_out["duration_target"],
            "energy_prediction": variance_adaptor_out["energy_prediction"],
            "energy_target": variance_adaptor_out["energy_target"],
            "pitch_prediction": variance_adaptor_out["pitch_prediction"],
            "pitch_target": variance_adaptor_out["pitch_target"],
        }

    def on_load_checkpoint(self, checkpoint):
        """Deserialize the checkpoint hyperparameters.
        Note, this shouldn't fail on different versions of pydantic anymore,
        but it will fail on breaking changes to the config. We should catch those exceptions
        and handle them appropriately."""
        self.config = FeaturePredictionConfig(
            **checkpoint["hyper_parameters"]["config"]
        )
        if (
            "stats" in checkpoint["hyper_parameters"]
            and checkpoint["hyper_parameters"]["stats"] is not None
        ):
            self.stats = Stats(**checkpoint["hyper_parameters"]["stats"])

    def on_save_checkpoint(self, checkpoint):
        """Serialize the checkpoint hyperparameters"""
        # Convert the config to a checkpoint-safe config
        checkpoint["hyper_parameters"]["config"] = self.config.model_checkpoint_dump()
        if self.stats is not None:
            checkpoint["hyper_parameters"]["stats"] = self.stats.model_dump(mode="json")

    def predict_step(self, batch, batch_idx):
        with torch.no_grad():
            return self(batch, inference=True)

    def training_step(self, batch, batch_idx):
        output = self(batch)
        losses = self.loss(output, batch, self.current_epoch)
        self.log_dict(
            {f"training/{k}_loss": v.item() for k, v in losses.items()}, prog_bar=True
        )
        return losses["total"]

    def _validation_global_step_0(self, batch, batch_idx) -> None:
        if self.global_step != 0:
            return

        assert self.logger is not None
        audio, _ = torchaudio.load(
            (
                Path(self.config.preprocessing.save_dir)
                / "audio"
                / "--".join(
                    [
                        batch["basename"][0],
                        batch["speaker"][0],
                        batch["language"][0],
                        f"audio-{self.config.preprocessing.audio.input_sampling_rate}.wav",
                    ]
                )
            ).__str__()
        )

        audio = audio.squeeze()
        # Log ground truth audio
        self.logger.experiment.add_audio(
            f"gt/wav_{batch['basename'][0]}",
            audio,
            self.global_step,
            self.config.preprocessing.audio.output_sampling_rate,
        )
        if self.config.training.vocoder_path:
            input_ = batch["mel"]
            vocoder_ckpt = torch.load(
                self.config.training.vocoder_path, map_location=input_.device
            )
            vocoder_model, vocoder_config = load_hifigan_from_checkpoint(
                vocoder_ckpt, input_.device
            )
            wav, sr = synthesize_data(input_, vocoder_model, vocoder_config)

            self.logger.experiment.add_audio(
                f"copy-synthesis/wav_{batch['basename'][0]}",
                wav,
                self.global_step,
                sr,
            )

    def _validation_batch_idx_0(self, batch, batch_idx, output) -> None:
        if batch_idx != 0:
            return

        assert self.logger is not None

        # Currently only plots the first one, but the function is writte to support plotting multiple
        if self.config.model.learn_alignment:
            figs = plot_attn_maps(
                output["attn_soft"],
                output["attn_hard"],
                output["tgt_lens"],
                output["src_lens"],
                n=1,
            )
            for i, fig in enumerate(figs):
                self.logger.experiment.add_figure(
                    f"attention/{batch['basename'][i]}", fig, self.global_step
                )

        duration_np = output["duration_target"][0].cpu().numpy()
        gt_pitch_for_plotting = batch["pitch"][0].cpu().numpy()
        gt_energy_for_plotting = batch["energy"][0].cpu().numpy()
        pred_pitch_for_plotting = output["pitch_prediction"][0].cpu().numpy()
        pred_energy_for_plotting = output["energy_prediction"][0].cpu().numpy()

        if self.config.model.variance_predictors.pitch.level == "phone":
            pred_pitch_for_plotting = expand(pred_pitch_for_plotting, duration_np)
            if not self.config.model.learn_alignment:
                # pitch targets are frame-wise if alignment is learned
                gt_pitch_for_plotting = expand(gt_pitch_for_plotting, duration_np)

        if self.config.model.variance_predictors.energy.level == "phone":
            pred_energy_for_plotting = expand(pred_energy_for_plotting, duration_np)
            if not self.config.model.learn_alignment:
                # energy targets are frame-wise if alignment is learned
                gt_energy_for_plotting = expand(gt_energy_for_plotting, duration_np)
        if self.stats is not None:
            self.logger.experiment.add_figure(
                f"pred/spec_{batch['basename'][0]}",
                plot_mel(
                    [
                        {
                            "mel": np.swapaxes(batch["mel"][0].cpu().numpy(), 0, 1),
                            "pitch": gt_pitch_for_plotting,
                            "energy": gt_energy_for_plotting,
                        },
                        {
                            "mel": np.swapaxes(
                                output[self.output_key][0].cpu().numpy(), 0, 1
                            ),
                            "pitch": pred_pitch_for_plotting,
                            "energy": pred_energy_for_plotting,
                        },
                    ],
                    self.stats,
                    ["Ground-Truth Spectrogram", "Synthesized Spectrogram"],
                ),
                self.global_step,
            )

        if self.config.training.vocoder_path:
            input_ = output[self.output_key]
            vocoder_ckpt = torch.load(
                self.config.training.vocoder_path, map_location=input_.device
            )
            vocoder_model, vocoder_config = load_hifigan_from_checkpoint(
                vocoder_ckpt, input_.device
            )
            wav, sr = synthesize_data(input_, vocoder_model, vocoder_config)
            self.logger.experiment.add_audio(
                f"pred/wav_{batch['basename'][0]}", wav, self.global_step, sr
            )

    def validation_step(self, batch, batch_idx):
        if self.global_step == 0:
            self._validation_global_step_0(batch, batch_idx)

        output = self(batch)
        if batch_idx == 0:
            self._validation_batch_idx_0(batch, batch_idx, output)

        losses = self.loss(output, batch, self.current_epoch)
        self.log_dict(
            {f"validation/{k}_loss": v.item() for k, v in losses.items()},
            batch_size=self.batch_size,
            sync_dist=True,
        )

    def configure_optimizers(self):
        self.optimizer = torch.optim.AdamW(
            self.parameters(),
            self.config.training.optimizer.learning_rate,
            betas=self.config.training.optimizer.betas,
            eps=self.config.training.optimizer.eps,
            weight_decay=self.config.training.optimizer.weight_decay,
        )

        self.scheduler = NoamLR(
            self.optimizer,
            self.config.training.optimizer.warmup_steps,
        )

        sched = {
            "scheduler": self.scheduler,
            "interval": "step",
        }

        return [self.optimizer], [sched]



""" Setup for everyvoice
"""

from os import path

from setuptools import find_packages, setup

# Ugly hack to read the current version number without importing everyvoice:
with open("everyvoice/_version.py", "r", encoding="utf8") as version_file:
    namespace: dict = {}
    exec(version_file.read(), namespace)
    VERSION = namespace["VERSION"]
    # [N!]N(.N)*[{a|b|rc}N][.postN][.devN]

this_directory = path.abspath(path.dirname(__file__))

with open(path.join(this_directory, "README.md"), encoding="utf8") as f:
    long_description = f.read()

with open(path.join(this_directory, "requirements.txt"), encoding="utf8") as f:
    REQS = f.read().splitlines()

# Get a list of submodules
SUBMODULE_PATHS = [
    "everyvoice/model/feature_prediction/FastSpeech2_lightning/fs2",
    "everyvoice/model/vocoder/HiFiGAN_iSTFT_lightning/hfgl",
    "everyvoice/model/aligner/DeepForcedAligner/dfaligner",
    "everyvoice/model/aligner/wav2vec2aligner/aligner",
]
SUBMODULE_PACKAGES = []

# For each submodule
for submodule_path in SUBMODULE_PATHS:
    # append the submodule path to the list of submodule packages
    submodule = submodule_path.replace("/", ".")
    SUBMODULE_PACKAGES.append(submodule)
    # then use find_packages to automatically find the rest
    packages = find_packages(submodule_path)
    SUBMODULE_PACKAGES += [submodule + f".{pkg}" for pkg in packages]

setup(
    name="everyvoice",
    python_requires=">=3.10, <3.12",
    version=VERSION,
    author="Aidan Pine",
    author_email="hello@aidanpine.ca",
    license="MIT",
    url="https://github.com/roedoejet/EveryVoice",
    description="Text-to-Speech Synthesis for the Speech Generation for Indigenous Language Education Small Teams Project",
    long_description=long_description,
    long_description_content_type="text/markdown",
    platform=["any"],
    packages=find_packages() + SUBMODULE_PACKAGES,
    include_package_data=True,
    install_requires=REQS,
    entry_points={"console_scripts": ["everyvoice = everyvoice.cli:app"]},
    zip_safe=False,
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "License :: OSI Approved :: MIT License",
        "Operating System :: POSIX :: Linux",
        "Operating System :: MacOS",
    ],
)



"""
If you've installed `everyvoice` and would like to run this unittest:
python -m unittest everyvoice.model.feature_prediction.FastSpeech2_lightning.fs2.tests.test_cli
"""

import io
from contextlib import redirect_stderr
from pathlib import Path
from tempfile import TemporaryDirectory
from unittest import TestCase

from everyvoice.config.shared_types import ContactInformation
from everyvoice.config.type_definitions import (
    DatasetTextRepresentation,
    TargetTrainingTextRepresentationLevel,
)
from everyvoice.tests.stubs import mute_logger
from everyvoice.utils import generic_psv_filelist_reader
from typer.testing import CliRunner

from ..cli.cli import app
from ..cli.synthesize import prepare_data as prepare_synthesize_data
from ..cli.synthesize import validate_data_keys_with_model_keys
from ..config import FastSpeech2Config

DEFAULT_LANG2ID: set = set()
DEFAULT_SPEAKER2ID: set = set()
CONTACT = ContactInformation(
    contact_name="Test Runner", contact_email="info@everyvoice.ca"
)


class SynthesizeTest(TestCase):
    def setUp(self) -> None:
        self.runner = CliRunner()

    def test_help(self):
        result = self.runner.invoke(app, ["synthesize", "--help"])
        self.assertIn("synthesize [OPTIONS] MODEL_PATH", result.stdout)

    def test_no_model(self):
        result = self.runner.invoke(app, ["synthesize"])
        self.assertIn("Missing argument 'MODEL_PATH'.", result.stdout)

    def test_filelist_and_text(self):
        with TemporaryDirectory() as tmpdir:
            tmpdir = Path(tmpdir)
            test = tmpdir / "test.psv"
            test.touch()
            model = tmpdir / "model"
            model.touch()
            result = self.runner.invoke(
                app,
                (
                    "synthesize",
                    "--filelist",
                    str(test),
                    "--text",
                    "BAD",
                    str(model),
                ),
            )
            self.assertIn(
                "Got arguments for both text and a filelist - this will only process the text."
                " Please re-run without providing text if you want to run batch synthesis",
                result.stdout,
            )

    def test_no_filelist_nor_text(self):
        with TemporaryDirectory() as tmpdir:
            tmpdir = Path(tmpdir)
            model = tmpdir / "model"
            model.touch()
            result = self.runner.invoke(
                app,
                (
                    "synthesize",
                    str(model),
                ),
            )
            self.assertIn("You must define either --text or --filelist", result.stdout)


class MockModelForPrepare:
    class Dummy:
        pass

    def __init__(
        self,
        lang2id,
        speaker2id,
        filelist_loader,
        multilingual,
        multispeaker,
        target_text_representation_level,
    ):
        self.lang2id = lang2id
        self.speaker2id = speaker2id
        self.config = self.Dummy()
        self.config.training = self.Dummy()
        self.config.training.filelist_loader = filelist_loader
        self.config.model = self.Dummy()
        self.config.model.multilingual = multilingual
        self.config.model.multispeaker = multispeaker
        self.config.model.target_text_representation_level = (
            target_text_representation_level
        )


class PrepareSynthesizeDataTest(TestCase):
    """"""

    def test_filelist_language(self):
        """
        Use a different language than the one provided in the filelist.
        """
        data = prepare_synthesize_data(
            texts=[],
            language="foo",
            speaker="bar",
            duration_control=1.0,
            filelist=Path(__file__).parent / "data/filelist.psv",
            model=MockModelForPrepare(
                lang2id={"foo": 1},
                speaker2id={"bar": 2},
                filelist_loader=generic_psv_filelist_reader,
                multilingual=True,
                multispeaker=True,
                target_text_representation_level=TargetTrainingTextRepresentationLevel.characters,
            ),
            text_representation=DatasetTextRepresentation.characters,
        )
        self.assertEqual(len(data), 10)
        self.assertTrue(all((d["language"] == "foo" for d in data)))

    def test_filelist_speaker(self):
        """
        Use a different speaker than the one provided in the filelist.
        """
        data = prepare_synthesize_data(
            texts=[],
            language="foo",
            speaker="bar",
            duration_control=1.0,
            filelist=Path(__file__).parent / "data/filelist.psv",
            model=MockModelForPrepare(
                lang2id={"foo": 1},
                speaker2id={"bar": 2},
                filelist_loader=generic_psv_filelist_reader,
                multilingual=True,
                multispeaker=True,
                target_text_representation_level=TargetTrainingTextRepresentationLevel.characters,
            ),
            text_representation=DatasetTextRepresentation.characters,
        )
        self.assertEqual(
            data[-1]["basename"],
            "LJ002 this is a really long basename",
            "Asserts that if a filelist provides a basename, it won't get slugified or truncated",
        )
        self.assertEqual(len(data), 10)
        self.assertTrue(all((d["speaker"] == "bar" for d in data)))

    def test_plain_filelist(self):
        with mute_logger("fs2.cli"):
            data = prepare_synthesize_data(
                texts=[],
                language=None,
                speaker=None,
                duration_control=1.0,
                filelist=Path(__file__).parent / "data/filelist.txt",
                model=MockModelForPrepare(
                    lang2id={"foo": 1},
                    speaker2id={"bar": 2},
                    filelist_loader=generic_psv_filelist_reader,
                    multilingual=True,
                    multispeaker=True,
                    target_text_representation_level=TargetTrainingTextRepresentationLevel.characters,
                ),
                text_representation=DatasetTextRepresentation.characters,
            )
        self.assertEqual(
            data[-1]["basename"],
            "Other-cases-are-reco-674e00ab",
            "Asserts that basenames are truncated and slugified",
        )
        self.assertEqual(len(data), 9)
        self.assertTrue(all((d["language"] == "foo" for d in data)))
        self.assertTrue(all((d["speaker"] == "bar" for d in data)))


class ValidateDataWithModelTest(TestCase):
    """
    Validate different combination of user provided options against model configuration.
    """

    def test_multilingual_invalid_language(self):
        """
        The model is multilingual and the user provided a language that is not supported by the model.
        """
        language = "UNSUPPORTED"
        config = FastSpeech2Config(contact=CONTACT)
        config.model.multilingual = True
        model_languages = {"L1", "L2"}
        f = io.StringIO()
        with self.assertRaises(SystemExit), redirect_stderr(f):
            validate_data_keys_with_model_keys(
                data_keys={language},
                model_keys=model_languages,
                key="language",
                multi=bool(model_languages),
            )
        self.assertIn(
            f"You provided {set((language,))} which is not a language supported by the model {model_languages}.",
            f.getvalue(),
        )
        language_two = "ALSO_UNSUPPORTED"
        with self.assertRaises(SystemExit), redirect_stderr(f):
            validate_data_keys_with_model_keys(
                data_keys={language, language_two},
                model_keys=model_languages,
                key="language",
                multi=bool(model_languages),
            )
        self.assertIn(
            f"You provided {set((language,language_two))} which are not languages that are supported by the model {model_languages}.",
            f.getvalue(),
        )

    def test_not_multilingual_with_language(self):
        """
        The model is not multilingual and the user provided a language.
        """
        language = "L3"
        config = FastSpeech2Config(contact=CONTACT)
        config.model.multilingual = False
        model_languages = DEFAULT_LANG2ID
        f = io.StringIO()
        with self.assertRaises(SystemExit), redirect_stderr(f):
            validate_data_keys_with_model_keys(
                data_keys={language},
                model_keys=model_languages,
                key="language",
                multi=bool(model_languages),
            )
        self.assertIn(
            "The current model doesn't support multiple languages",
            f.getvalue(),
        )

    def test_multispeaker_invalid_speaker(self):
        """
        The model is multispeaker and the user provided a speaker that is not supported by the model.
        """
        speaker = "UNSUPPORTED"
        config = FastSpeech2Config(contact=CONTACT)
        config.model.multispeaker = True
        model_speakers = {"S1", "S2"}
        f = io.StringIO()
        with self.assertRaises(SystemExit), redirect_stderr(f):
            validate_data_keys_with_model_keys(
                data_keys={speaker},
                model_keys=model_speakers,
                key="speaker",
                multi=bool(model_speakers),
            )
        self.assertIn(
            f"You provided {set((speaker,))} which is not a speaker supported by the model {model_speakers}.",
            f.getvalue(),
        )

    def test_not_multispeaker_with_speaker(self):
        """
        The model is not multispeaker and the user provided a speaker.
        """
        speaker = "s3"
        config = FastSpeech2Config(contact=CONTACT)
        config.model.multispeaker = False
        model_speakers = DEFAULT_SPEAKER2ID
        f = io.StringIO()
        with self.assertRaises(SystemExit), redirect_stderr(f):
            validate_data_keys_with_model_keys(
                data_keys={speaker},
                model_keys=model_speakers,
                key="speaker",
                multi=bool(model_speakers),
            )
        self.assertIn(
            "The current model doesn't support multiple speakers",
            f.getvalue(),
        )


class CLITest(TestCase):
    """
    Validate that all subcommands are accessible.
    """

    def setUp(self) -> None:
        self.runner = CliRunner()
        self.subcommands = (
            "audit",
            "benchmark",
            "check_data",
            "preprocess",
            "synthesize",
            "train",
        )

    def test_commands_present(self):
        """
        Each subcommand is present in the the command's help message.
        """
        result = self.runner.invoke(app, ["--help"])
        for command in self.subcommands:
            with self.subTest(msg=f"Looking for {command}"):
                self.assertIn(command, result.stdout)

    def test_command_help_messages(self):
        """
        Each subcommand has its help message.
        """
        for subcommand in self.subcommands:
            with self.subTest(msg=f"Looking for {subcommand}'s help"):
                result = self.runner.invoke(app, [subcommand, "--help"])
                self.assertEqual(result.exit_code, 0)
                result = self.runner.invoke(app, [subcommand, "-h"])
                self.assertEqual(result.exit_code, 0)



import os
from pathlib import Path

import numpy as np
import pytorch_lightning as pl
import torch
import torch.nn as nn
from everyvoice.model.aligner.config import AlignerConfig
from everyvoice.text.text_processor import TextProcessor

from .config import DFAlignerConfig
from .duration_extraction import extract_durations_with_dijkstra


class BatchNormConv(nn.Module):
    def __init__(self, in_channels: int, out_channels: int, kernel_size: int):
        super().__init__()
        self.conv = nn.Conv1d(
            in_channels,
            out_channels,
            kernel_size,
            stride=1,
            padding=kernel_size // 2,
            bias=False,
        )
        self.bnorm = nn.BatchNorm1d(out_channels)
        self.relu = nn.ReLU()

    def forward(self, x):
        x = x.transpose(1, 2)
        x = self.conv(x)
        x = self.relu(x)
        x = self.bnorm(x)
        x = x.transpose(1, 2)
        return x


class Aligner(pl.LightningModule):
    def __init__(
        self,
        config: dict | DFAlignerConfig,
    ) -> None:
        super().__init__()
        if isinstance(config, dict):
            config = AlignerConfig(**config)
        self.config: AlignerConfig = config  # type: ignore
        self.preprocessed_dir = Path(self.config.preprocessing.save_dir)
        self.sep = "--"
        self.text_processor = TextProcessor(self.config.text)
        conv_dim = self.config.model.conv_dim
        lstm_dim = self.config.model.lstm_dim
        n_mels = self.config.preprocessing.audio.n_mels
        num_symbols = len(self.text_processor.symbols)
        self.log_dir = os.path.join(
            self.config.training.logger.save_dir,
            self.config.training.logger.name,
        )
        self.longest_mel = None
        self.longest_tokens = None
        self.batch_size = (
            self.config.training.batch_size
        )  # this is declared explicitly so that auto_scale_batch_size works: https://pytorch-lightning.readthedocs.io/en/stable/advanced/training_tricks.html
        self.register_buffer("step", torch.tensor(1, dtype=torch.int))
        self.ctc_loss = nn.CTCLoss()
        self.save_hyperparameters()  # TODO: ignore=['specific keys'] - I should ignore some unnecessary/problem values
        self.convs = nn.ModuleList(
            [
                BatchNormConv(n_mels, conv_dim, 5),
                BatchNormConv(conv_dim, conv_dim, 5),
                BatchNormConv(conv_dim, conv_dim, 5),
            ]
        )
        self.rnn = torch.nn.LSTM(
            conv_dim, lstm_dim, batch_first=True, bidirectional=True
        )
        self.lin = torch.nn.Linear(2 * lstm_dim, num_symbols)

    def forward(self, x):
        if self.train:
            self.step += 1
        for conv in self.convs:
            x = conv(x)
        x, _ = self.rnn(x)
        x = self.lin(x)
        return x

    def on_load_checkpoint(self, checkpoint):
        """Deserialize the checkpoint hyperparameters.
        Note, this shouldn't fail on different versions of pydantic anymore,
        but it will fail on breaking changes to the config. We should catch those exceptions
        and handle them appropriately."""
        self.config = AlignerConfig(**checkpoint["hyper_parameters"]["config"])

    def on_save_checkpoint(self, checkpoint):
        """Serialize the checkpoint hyperparameters"""
        checkpoint["hyper_parameters"]["config"] = self.config.model_checkpoint_dump()

    def configure_optimizers(self):
        optim = torch.optim.AdamW(
            self.parameters(), self.config.training.optimizer.learning_rate
        )
        scheduler = torch.optim.lr_scheduler.ReduceLROnPlateau(optim)
        return {
            "optimizer": optim,
            "lr_scheduler": scheduler,
            "monitor": "validation/loss",
        }

    def predict_step(self, batch, batch_idx):
        save_dir = Path(self.config.preprocessing.save_dir)
        sep = "--"
        tokens = batch["tokens"]
        mel = batch["mel"]
        mel_len = batch["mel_len"]
        pred_batch = self(mel)
        for b in range(tokens.size(0)):
            this_mel_len = mel_len[b]
            pred = pred_batch[b, :this_mel_len, :]
            pred = torch.softmax(pred, dim=-1)
            pred = pred.detach().cpu().numpy()
            basename = batch["basename"][b]
            speaker = batch["speaker"][b]
            language = batch["language"][b]
            np.save(
                save_dir
                / "duration"
                / sep.join([basename, speaker, language, "duration.npy"]),
                pred,
                allow_pickle=False,
            )

    def training_step(self, batch, batch_idx):
        loss = self._calculate_loss_from_batch(batch)
        self.log("training/loss", loss.item(), prog_bar=False)
        return loss

    def validation_step(self, batch, batch_idx):
        loss = self._calculate_loss_from_batch(batch)
        self.log("validation/loss", loss.item(), prog_bar=False)
        if batch_idx == 0:
            self._generate_plots(batch["mel"], batch["tokens"])
        return loss

    def _calculate_loss_from_batch(self, batch):
        tokens = batch["tokens"]
        mel = batch["mel"]
        tokens_len = batch["tokens_len"]
        mel_len = batch["mel_len"]
        pred = self(mel)
        pred = pred.transpose(0, 1).log_softmax(2)
        return self.ctc_loss(pred, tokens, mel_len, tokens_len)

    def _generate_plots(self, mel, tokens):
        if self.longest_mel is None:
            self.longest_mel = mel
            self.longest_tokens = tokens.detach().cpu()
        pred = self(self.longest_mel)[0].detach().cpu().softmax(dim=-1)
        durations = extract_durations_with_dijkstra(
            self.longest_tokens.squeeze(0).numpy(), pred.numpy()
        )
        pred_max = pred.max(1)[1].numpy().tolist()
        pred_text = self.text_processor.decode_tokens(pred_max)
        target_text = self.text_processor.decode_tokens(
            self.longest_tokens.squeeze().tolist()
        )
        target_duration_rep = "".join(
            c * durations[i]
            for i, c in enumerate(self.text_processor._tokenizer.tokenize(target_text))
        )
        tensorboard = self.logger.experiment
        tensorboard.add_text(
            "validation/prediction", f"    {pred_text}", global_step=self.global_step
        )
        tensorboard.add_text(
            "validation/target", f"    {target_text}", global_step=self.global_step
        )
        tensorboard.add_text(
            "validation/target_duration_rep",
            f"    {target_duration_rep}",
            global_step=self.global_step,
        )



"""
Utility functions that require expensive imports. Don't import from cli until needed.
"""

import re

import torch

# from torchaudio.models import wav2vec2_model
import torchaudio
from torchaudio.functional import forced_align

from .classes import Frame, Segment

DEVICE = torch.device("cuda" if torch.cuda.is_available() else "cpu")


def load_model():
    bundle = torchaudio.pipelines.WAV2VEC2_ASR_LARGE_960H
    model = bundle.get_model()
    labels = {l.lower(): i for i, l in enumerate(bundle.get_labels())}
    del labels["-"]  # Remove blank token
    del labels["|"]  # Remove sentence token
    model.to(DEVICE)
    return model, labels


def align_speech_file(
    audio, text_hash, model, labels_dictionary, word_padding, sentence_padding
):
    emission = get_emission(model, audio.to(DEVICE))
    segments, words, sentences = compute_alignments(
        text_hash,
        labels_dictionary,
        emission,
        word_padding=word_padding,
        sentence_padding=sentence_padding,
    )
    return segments, words, sentences, emission.size(1)


def get_emission(model, waveform):
    with torch.inference_mode():
        # NOTE: this step is essential
        waveform = torch.nn.functional.layer_norm(waveform, waveform.shape)
        emission, _ = model(waveform)
        return torch.log_softmax(emission, dim=-1)


def compute_alignments(
    transcript_hash, dictionary, emission, word_padding=0, sentence_padding=0
):
    all_words = [
        v["text"].output_string for k, v in transcript_hash.items() if "w" in k
    ]
    transcript = "".join(all_words)
    tokens = [dictionary[c] for c in transcript]
    targets = torch.tensor([tokens], dtype=torch.int32, device=emission.device)
    input_lengths = torch.tensor([emission.shape[1]], device=emission.device)
    target_lengths = torch.tensor([targets.shape[1]], device=emission.device)

    alignment, scores = forced_align(
        emission, targets, input_lengths, target_lengths, 0
    )

    scores = scores.exp()  # convert back to probability
    alignment, scores = alignment[0].tolist(), scores[0].tolist()

    assert len(alignment) == len(scores) == emission.size(1)
    token_index = -1
    prev_hyp = 0
    frames = []
    for i, (ali, score) in enumerate(zip(alignment, scores)):
        if ali == 0:
            prev_hyp = 0
            continue

        if ali != prev_hyp:
            token_index += 1
        frames.append(Frame(token_index, i, score))
        prev_hyp = ali
    words_to_match = [v | {"key": k} for k, v in transcript_hash.items() if "w" in k]
    i1, i2 = 0, 0
    segments = []
    while i1 < len(frames):
        while i2 < len(frames) and frames[i1].token_index == frames[i2].token_index:
            i2 += 1
        score = sum(frames[k].score for k in range(i1, i2)) / (i2 - i1)

        segments.append(
            Segment(
                transcript[frames[i1].token_index],
                frames[i1].time_index,
                frames[i2 - 1].time_index + 1,
                score,
            )
        )
        i1 = i2
    segments_to_match = segments.copy()
    while len(words_to_match) > 0:
        current_word = words_to_match.pop(0)
        current_segment_sequence = ""
        scores = []

        if segments_to_match:
            start = segments_to_match[0].start
        end = None

        while len(segments_to_match) > 0:
            current_segment = segments_to_match.pop(0)
            scores.append(current_segment.score)
            end = current_segment.end

            current_segment_sequence += current_segment.label
            if (
                current_segment_sequence
                == current_word[
                    "text"
                ].output_string  # break the loop if the word is equal to the output string
            ):
                break
        if end is not None:
            transcript_hash[current_word["key"]] = Segment(
                current_word["text"].input_string,
                start - word_padding,
                end + word_padding,
                sum(scores) / len(scores),
            )

    key_pattern = re.compile(
        r"""
            (s\d+)                   # sentence
            (w\d+)                   # word
             """,
        re.VERBOSE | re.IGNORECASE,
    )
    word_hash = {k: v for k, v in transcript_hash.items() if "w" in k}
    for sentence in [k for k in transcript_hash.keys() if "w" not in k]:
        scores = []
        start = None
        end = None
        for w_k, w_v in word_hash.items():
            if sentence == re.match(key_pattern, w_k).group(1):
                scores.append(w_v.score)
                if start is None:
                    start = w_v.start
                end = w_v.end
        if end is not None:
            transcript_hash[sentence] = Segment(
                transcript_hash[sentence]["text"].input_string,
                start - sentence_padding,
                end + sentence_padding,
                sum(scores) / len(scores),
            )
    words = [v for k, v in transcript_hash.items() if "w" in k]
    sentences = [v for k, v in transcript_hash.items() if "w" not in k]
    return segments, words, sentences



import re
from collections import Counter
from typing import Optional, Type, overload

import numpy as np
import numpy.typing as npt
from loguru import logger
from nltk.tokenize import RegexpTokenizer

from everyvoice.config.text_config import TextConfig
from everyvoice.exceptions import OutOfVocabularySymbolError
from everyvoice.text.features import PhonologicalFeatureCalculator
from everyvoice.text.phonemizer import AVAILABLE_G2P_ENGINES, get_g2p_engine
from everyvoice.text.utils import (
    apply_cleaners_helper,
    apply_to_replace_helper,
    normalize_text_helper,
)

PAD_SYMBOL = "\x80"


class TextProcessor:
    """Text is processed like:

    InputText (str, either DatasetTextRepresentation.characters, DatasetTextRepresentation.ipa_phones, or DatasetTextRepresentation.arpabet)
      -> to_replace operations
        -> cleaner operations
        = Cleaned Text (str)
          -> Optional[grapheme-to-phoneme, outputs tokens] OR
          -> tokenization
          = Tokens (list[str])
            -> Punctuation mapped to internal representation
            = Tokens with internal punctuation representation (list[str])
    """

    def __init__(self, config: TextConfig):
        self.config = config
        self.phonological_feature_calculator: Optional[
            PhonologicalFeatureCalculator
        ] = None
        self._pad_symbol = PAD_SYMBOL  # Use the Unicode PAD symbol

        # Add punctuation
        # Add an internal hash to convert from the type of Punctuation to the internal representation
        self.punctuation_internal_hash = {
            "exclamations": "<EXCL>",
            "question_symbols": "<QINT>",
            "quotemarks": "<QUOTE>",
            "big_breaks": "<BB>",
            "small_breaks": "<SB>",
            "ellipsis": "<EPS>",
        }
        # Create a hash table from punctuation to the internal ID
        self.punctuation_to_internal_id = {
            v: self.punctuation_internal_hash[punctuation_type]
            for punctuation_type, punctuation_type_values in iter(
                self.config.symbols.punctuation
            )
            for v in punctuation_type_values
        }
        self.punctuation_characters = list(self.punctuation_to_internal_id.keys())
        assert set(self.punctuation_characters) == self.config.symbols.punctuation.all

        # Add the internal punctuation IDs to the symbols list
        # Combine all the symbol fields into one list (except for punctuation)
        symbols = self.config.symbols.all_except_punctuation
        symbols |= set(self.punctuation_internal_hash.values())
        symbols |= self.config.symbols.punctuation.all
        # TODO: do I need to clean the symbols? How to do this if datasets have
        #       their own cleaners?
        _hardcoded_internal_symbols = [self._pad_symbol, " "]
        self.symbols = _hardcoded_internal_symbols + list(
            sorted(
                # Remove duplicates from symbol list, and apply longest
                # characters first to apply multigraph symbols first
                symbols - set(_hardcoded_internal_symbols),
                key=lambda symbol: (
                    -len(symbol),
                    symbol,
                ),  # reverse-length sort, then sort alphabetically
            )
        )
        self.to_replace = config.to_replace
        self.missing_symbols: Counter[str] = Counter()

        # Mappings from symbol to numeric ID and vice versa
        # dicts are preferred here to lists although list[index]
        # and list.index(symbol) could provide the same result
        # dicts are redundant protection against duplicate symbols
        self._symbol_to_id: dict[str, int] = {}
        self._id_to_symbol: dict[int, str] = {}
        for i, s in enumerate(self.symbols):
            self._symbol_to_id[s] = i
            self._id_to_symbol[i] = s

        self._tokenizer = RegexpTokenizer(
            "|".join([re.escape(x) for x in self.symbols + self.punctuation_characters])
        )
        self._missing_symbol_finder = RegexpTokenizer(
            "|".join(
                [re.escape(x) for x in self.symbols + self.punctuation_characters]
            ),
            gaps=True,
            discard_empty=True,
        )

    def get_missing_symbols(
        self, text: str, normalize_text=True, quiet=False
    ) -> list[str]:
        """Helper function to return a list of symbols missing from configuration.

        Args:
            text (str): text to find missing symbols in
            normalize_text (bool, optional): whether to normalize text first. Defaults to True.

        Returns:
            list[str]: a list of missing symbols in the text. globs all adjacent missing symbols together

        >>> tp = TextProcessor(TextConfig())
        >>> tp.get_missing_symbols(' ç **', quiet=True)
        ['ç', '**']
        """
        if normalize_text:
            text = self.normalize_text(text)
        for symbol in (missing_tokens := self._missing_symbol_finder.tokenize(text)):
            if not quiet:
                logger.warning(
                    f"Symbol '{symbol}' occurs in the text '{text}' but was not declared in your configuration so it is being ignored."
                )
            self.missing_symbols[symbol] += 1
        return missing_tokens

    def apply_punctuation_rules(self, tokenized_text: list[str]) -> list[str]:
        """Given some text, normalize all punctuation according to internal representation

        Args:
            text (list[str]): tokenized text with punctuation

        Returns:
            list[str]: tokenized text with normalized punctuation

        >>> tp = TextProcessor(TextConfig())
        >>> tp.apply_punctuation_rules(['h', 'e', 'l', 'l', 'o', '.'])
        ['h', 'e', 'l', 'l', 'o', '<BB>']
        """
        return [
            self.punctuation_to_internal_id.get(token, token)
            for token in tokenized_text
        ]

    def apply_replacement_rules(self, text: str) -> str:
        """Given some text and a list of replacement operations in the form of input/output key value pairs,
           return the transformed text.
        Args:
            text (str): The text to be converted
        Returns:
            str: the replaced text

        >>> tp = TextProcessor(TextConfig(to_replace={'a': 'b'}))
        >>> tp.apply_replacement_rules('a')
        'b'
        """
        return apply_to_replace_helper(text, self.to_replace)

    def apply_cleaners(self, text: str) -> str:
        """Converts some text to cleaned text

        Args:
            text (str): The text to be converted
        Returns:
            str: the replaced text

        >>> from everyvoice.utils import collapse_whitespace, lower, nfc_normalize
        >>> tp = TextProcessor(TextConfig(cleaners=[collapse_whitespace, lower, nfc_normalize]))
        >>> tp.apply_cleaners('HELLO\u0301')
        'helló'
        """
        return apply_cleaners_helper(text, self.config.cleaners)

    def normalize_text(
        self, text: str, apply_replace_rules=True, apply_cleaners=True
    ) -> str:
        """Normalize text by applying replace rules and all defined cleaners

        Args:
            text (str): un-normalized text
            apply_replace_rules (bool, optional): Whether to apply replace rules. Defaults to True.
            apply_cleaners (bool, optional): Whether to apply cleaners. Defaults to True.

        Returns:
            str: normalized text ready to be tokenized

        >>> from everyvoice.utils import collapse_whitespace, lower, nfc_normalize
        >>> tp = TextProcessor(TextConfig(cleaners=[collapse_whitespace, lower, nfc_normalize]))
        >>> tp.normalize_text('HELLO\u0301!')
        'helló!'
        """
        return normalize_text_helper(
            text,
            self.to_replace,
            self.config.cleaners,
            apply_cleaners=apply_cleaners,
            apply_replace_rules=apply_replace_rules,
        )

    def calculate_phonological_features(
        self, phone_tokens: list[str], apply_punctuation_rules=True
    ) -> npt.NDArray[np.float32]:
        """Converts a string of text to a sequence of IDs corresponding to the symbols in the text.
        Args:
            text (list[str]): a list of IPA and normalized punctuation tokens
            apply_punctuation_rules (bool): whether to convert punctuation into discrete values like <BB> or <EXCL>. Defaults to True. If set to False, features.py will have to be amended to handle punctuation.
        Returns:
            npt.NDArray[np.float32]: a list of multi-hot phonological feature vectors

        >>> tp = TextProcessor(TextConfig())
        >>> tp.calculate_phonological_features(['aɪ'])
        array([[ 1.,  1., -1.,  1., -1., -1., -1.,  0.,  1., -1., -1.,  0., -1.,
                 0., -1.,  0.,  0., -1., -1., -1.,  0., -1.,  0.,  0.,  0.,  0.,
                 0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.]],
              dtype=float32)
        """
        if self.phonological_feature_calculator is None:
            self.phonological_feature_calculator = PhonologicalFeatureCalculator(
                text_config=self.config,
                punctuation_hash=self.punctuation_internal_hash,
            )
        if apply_punctuation_rules:
            phone_tokens = self.apply_punctuation_rules(phone_tokens)
        return self.phonological_feature_calculator.get_features(phone_tokens)

    def apply_g2p_and_tokenization(
        self, normalized_text: str, lang_id: str, find_missing=True, quiet=False
    ) -> list[str]:
        """Converts a normalized string of graphemes for a particular language into a list of phone tokens.

        Args:
            normalized_text (str): a normalized string of graphemes
            lang_id (str): the language id

        Returns:
            list[str]: a list of phone tokens

        >>> from everyvoice.config.text_config import Symbols
        >>> tp = TextProcessor(TextConfig(symbols=Symbols(ipa=['a', 'h', 'ʌ', 'l', 'o', 'ʊ'])))
        >>> tp.apply_g2p_and_tokenization('hello', 'eng')
        ['h', 'ʌ', 'l', 'o', 'ʊ']
        """
        g2p_engine = get_g2p_engine(lang_id)
        try:
            tokens = g2p_engine(normalized_text)
        except Exception:
            tokens = None
            # TODO: do something here
            pass
        assert isinstance(
            tokens, list
        ), f"The g2p engine for {lang_id} produced {type(tokens)} but must produce a list of tokenized phones."
        valid_tokens = []
        punctuation_set = set(self.punctuation_characters)
        for token in tokens:
            if token in self._symbol_to_id or token in punctuation_set:
                valid_tokens.append(token)
            else:
                if find_missing:
                    if not quiet:
                        logger.warning(
                            f"Symbol '{token}' occurs in the text '{normalized_text}' but was not declared in your configuration so it is being ignored."
                        )
                self.missing_symbols[token] += 1
                continue

        return valid_tokens

    def apply_tokenization(
        self, normalized_text: str, quiet: bool = False, find_missing=True
    ) -> list[str]:
        """Converts a string of normalized text to a sequence of tokens.

        Args:
            text (str): string to convert to a sequence
            quiet (bool): suppress warnings
            find_missing (bool): find missing tokens and log them as part of the TextProcessor.missing_symbols counter
        Returns:
            list[str]: List of symbols in the text without missing symbols

        >>> tp = TextProcessor(TextConfig())
        >>> tp.apply_tokenization('\x80\x80 *', quiet=True)
        ['\x80', '\x80', ' ']
        >>> tp.missing_symbols['*']
        1
        """
        if find_missing:
            self.get_missing_symbols(normalized_text, quiet=quiet)
        tokens = self._tokenizer.tokenize(normalized_text)
        return tokens

    def encode_text(
        self,
        text: str,
        normalize_text: bool = True,
        apply_g2p: bool = False,
        normalize_punctuation: bool = False,
        encode_as_phonological_features: bool = False,
        lang_id: Optional[str] = None,
        quiet: bool = False,
        find_missing: bool = True,
    ) -> list[int] | npt.NDArray[np.float32]:
        """Converts a string of text to a sequence of IDs corresponding to the symbols in the text.
        Args:
            text: string to convert to a sequence
        Returns:
            list[int]|list[list[int]]: Either a one-hot encoding of integers
                corresponding to the symbols in the text, or a multi-hot
                phonological feature vector

        >>> from everyvoice.config.text_config import Symbols
        >>> tp = TextProcessor(TextConfig(symbols=Symbols(ipa=['a', 'h', 'ʌ', 'l', 'o', 'ʊ'])))
        >>> tp.encode_text('hello \x80\x80', quiet=True) # e is not in the default symbols so it is ignored
        [19, 20, 20, 21, 1, 0, 0]
        >>> tp.encode_text('hello \x80\x80', apply_g2p=True, lang_id='boop', quiet=True)
        Traceback (most recent call last):
        ...
        ValueError: You tried to apply g2p for language 'boop', but no g2p engine exists for that language. Please see the <TODO: docs>.
        >>> tp.encode_text('hello \x80\x80', apply_g2p=False, lang_id='boop', encode_as_phonological_features=True, quiet=True)
        Traceback (most recent call last):
        ...
        ValueError: 'encode_as_phonological_features' was set to True but 'apply_g2p' was set to False. In order to calculate phonological features, you must first apply g2p to the text. Please set 'apply_g2p' to True.
        >>> tp.encode_text('hello \x80\x80', apply_g2p=True, lang_id='eng', quiet=True)
        [19, 27, 20, 21, 26, 1, 0, 0]
        """
        # Error states
        if encode_as_phonological_features and not apply_g2p:
            raise ValueError(
                "'encode_as_phonological_features' was set to True but 'apply_g2p' was set to False."
                " In order to calculate phonological features, you must first apply g2p to the text."
                " Please set 'apply_g2p' to True."
            )
        if apply_g2p and (lang_id is None or lang_id not in AVAILABLE_G2P_ENGINES):
            raise ValueError(
                f"You tried to apply g2p for language '{lang_id}', but no g2p engine exists for that language."
                " Please see the <TODO: docs>."
            )

        if normalize_text:
            text = self.normalize_text(text)
        if apply_g2p and lang_id is not None:
            tokens = self.apply_g2p_and_tokenization(
                normalized_text=text,
                lang_id=lang_id,
                quiet=quiet,
                find_missing=find_missing,
            )
        else:
            tokens = self.apply_tokenization(
                text, quiet=quiet, find_missing=find_missing
            )
        if normalize_punctuation:
            tokens = self.apply_punctuation_rules(tokens)
        if encode_as_phonological_features:
            # applying punctuation rules will have already happened here, so set to False
            return self.calculate_phonological_features(
                tokens, apply_punctuation_rules=False
            )
        else:
            # TODO: catch errors
            return [self._symbol_to_id[symbol] for symbol in tokens]

    def token_sequence_to_text_sequence(self, sequence: list[int]) -> list[str]:
        """Converts a sequence of IDs to a sequence of text characters

        Args:
            sequence (list[int]): a sequence of IDs

        Returns:
            list[str]: a sequence of text characters

        >>> tp = TextProcessor(TextConfig())
        >>> tp.token_sequence_to_text_sequence([0, 6, 0, 0])
        ['\x80', '<SIL>', '\x80', '\x80']
        """
        return [self._id_to_symbol[symbol_id] for symbol_id in sequence]

    def encode_string_tokens(self, sequence: list[str]) -> list[int]:
        """Encode a sequence of string tokens

        Args:
            sequence (list[str]): a list of string tokens

        Returns:
            list[int]: a list of token indices

        >>> tp = TextProcessor(TextConfig())
        >>> tp.encode_string_tokens(['\x80', '<SIL>', '\x80', '\x80'])
        [0, 6, 0, 0]
        """
        # TODO: catch errors
        encoded_tokens = []
        for string_token in sequence:
            try:
                encoded_tokens.append(self._symbol_to_id[string_token])
            except KeyError as e:
                raise OutOfVocabularySymbolError(
                    f"Sequence {sequence} contains item {string_token}"
                ) from e
        return encoded_tokens

    def encode_escaped_string_sequence(
        self, string_of_tokens: str, split_character="/"
    ):
        assert (
            len(split_character) >= 1
        ), "An escaped string sequence must have a character to split on (default is '/')"
        return self.encode_string_tokens(
            [token for token in string_of_tokens.split(split_character) if token]
        )

    @overload
    def decode_tokens(  # noqa E704
        self, sequence: list[int], join_character: Type[None]
    ) -> list[str]: ...

    @overload
    def decode_tokens(  # noqa E704
        self, sequence: list[int], join_character: str
    ) -> str: ...

    def decode_tokens(self, sequence: list[int], join_character="/") -> str | list[str]:
        """Decode a sequence of encoded phone or character tokens into a sequence of strings

        Args:
            sequence (List[int]): sequence of phone or character tokens

        Returns:
            str: the string equivalent of the sequence

        >>> tp = TextProcessor(TextConfig())
        >>> tp.decode_tokens([0, 1, 2, 0, 0])
        '\x80/ /<QUOTE>/\x80/\x80'
        """
        if join_character is None:
            return self.token_sequence_to_text_sequence(sequence)
        else:
            return join_character.join(self.token_sequence_to_text_sequence(sequence))



import sys
from enum import Enum
from typing import Optional, Sequence

from anytree import NodeMixin, RenderTree

from .utils import EnumDict as State

TEXT_CONFIG_FILENAME_PREFIX = "everyvoice-shared-text"
ALIGNER_CONFIG_FILENAME_PREFIX = "everyvoice-aligner"
PREPROCESSING_CONFIG_FILENAME_PREFIX = "everyvoice-shared-data"
TEXT_TO_SPEC_CONFIG_FILENAME_PREFIX = "everyvoice-text-to-spec"
SPEC_TO_WAV_CONFIG_FILENAME_PREFIX = "everyvoice-spec-to-wav"
TEXT_TO_WAV_CONFIG_FILENAME_PREFIX = "everyvoice-text-to-wav"


class _Step:
    """The main class for steps within a tour.

    Each step must implement the prompt and validate methods.
    This method must save the answer in the response attribute.
    """

    def __init__(self, name: str):
        self.response = None
        self.completed = False
        self.name = name

    def prompt(self):
        """Prompt the user and return the result.

        Raises:
            NotImplementedError: If you don't implement anything, this will be raised
        """
        raise NotImplementedError(
            f"This step ({self.name}) doesn't have a prompt method implemented. Please implement one."
        )

    def sanitize_input(self, response):
        """
        Perform data sanitization of user provided input.
        """
        return response

    def validate(self, response) -> bool:
        """Validate the response.

        Raises:
            NotImplementedError: _description_
        """
        raise NotImplementedError(
            f"This step ({self.name}) doesn't have a validate method implemented. Please implement one."
        )

    def effect(self):
        """Run some arbitrary code after the step resolves"""
        pass


class Step(_Step, NodeMixin):
    """Just a mixin to allow a tree-based interpretation of steps"""

    def __init__(
        self,
        name: None | Enum | str = None,
        default=None,
        prompt_method=None,
        validate_method=None,
        effect_method=None,
        parent=None,
        children=None,
        state_subset=None,
    ):
        if name is None:
            name = getattr(self, "DEFAULT_NAME", "default step name missing")
        name = name.value if isinstance(name, Enum) else name
        super(Step, self).__init__(name)
        self.default = default
        self.parent = parent
        self.state_subset = state_subset
        # state will be added when the Step is added to a Tour
        self.state: Optional[State] = None
        # tour will be added when the Step is added to a Tour
        self.tour: Optional[Tour] = None
        if effect_method is not None:
            self.effect = effect_method  # type: ignore[method-assign]
        if prompt_method is not None:
            self.prompt = prompt_method  # type: ignore[method-assign]
        if validate_method is not None:
            self.validate = validate_method  # type: ignore[method-assign]
        if children:
            self.children = children
        self._validation_failures = 0

    def __repr__(self) -> str:
        return f"{self.name}: {super().__repr__()}"

    def run(self):
        """Prompt the user and save the response to the response attribute.
        If this method returns something truthy, continue, otherwise ask the prompt again.
        """
        self.response = self.prompt()
        self.response = self.sanitize_input(self.response)
        if self.validate(self.response):
            self.completed = True
            try:
                if self.state is not None:
                    self.state[self.name] = self.response
                return self.response
            finally:
                self.effect()
        else:
            self._validation_failures += 1
            if self._validation_failures > 20:
                print(f"ERROR: {self.name} giving up after 20 validation failures.")
                sys.exit(1)
            self.run()


class RootStep(Step):
    """Dummy step sitting at the root of the tour"""

    DEFAULT_NAME = "Root"

    def run(self):
        pass


class Tour:
    def __init__(self, name: str, steps: list[Step], state: Optional[State] = None):
        """Create the tour by placing all steps under a dummy root node"""
        self.name = name
        self.state: State = state if state is not None else State()
        self.steps = steps
        self.root = RootStep()
        self.root.tour = self
        self.determine_state(self.root, self.state)
        self.add_steps(steps, self.root)

    def determine_state(self, step: Step, state: State):
        if step.state_subset is not None:
            if step.state_subset not in state:
                state[step.state_subset] = State()
            step.state = state[step.state_subset]
        else:
            step.state = state

    def add_steps(self, steps: Sequence[Step | list[Step]], parent: Step):
        """Insert steps in front of the other children of parent.

        Steps are added as direct children.
        For sublists of steps, the first is a direct child, the rest are under it.
        """
        for item in reversed(steps):
            if isinstance(item, list):
                step, *children = item
                self.add_step(step, parent)
                self.add_steps(children, step)
            else:
                self.add_step(item, parent)

    def add_step(self, step: Step, parent: Step, child_index=0):
        self.determine_state(step, self.state)
        step.tour = self
        children = list(parent.children)
        children.insert(child_index, step)
        parent.children = children

    def run(self):
        for _, _, node in RenderTree(self.root):
            node.run()

    def visualize(self):
        for pre, _, node in RenderTree(self.root):
            treestr = f"{pre}{node.name}"
            print(treestr.ljust(8))


class StepNames(Enum):
    name_step = "Name Step"
    contact_name_step = "Contact Name Step"
    contact_email_step = "Contact Email Step"
    dataset_name_step = "Dataset Name Step"
    dataset_permission_step = "Dataset Permission Step"
    output_step = "Output Path Step"
    wavs_dir_step = "Wavs Dir Step"
    filelist_step = "Filelist Step"
    filelist_format_step = "Filelist Format Step"
    validate_wavs_step = "Validate Wavs Step"
    filelist_text_representation_step = "Filelist Text Representation Step"
    target_training_representation_step = (
        "Target Training Representation Step"  # TODO: maybe don't need
    )
    data_has_header_line_step = "Filelist Has Header Line Step"
    basename_header_step = "Basename Header Step"
    text_header_step = "Text Header Step"
    data_has_speaker_value_step = "Data Has Speaker Step"
    speaker_header_step = "Speaker Header Step"
    data_has_language_value_step = "Data Has Language Step"
    language_header_step = "Language Header Step"
    select_language_step = "Select Language Step"
    text_processing_step = "Text Processing Step"
    g2p_step = "G2P Step"
    symbol_set_step = "Symbol-Set step"
    sample_rate_config_step = "Sample Rate Config Step"
    audio_config_step = "Audio Config Step"
    sox_effects_step = "SoX Effects Step"
    more_datasets_step = "More Datasets Step"
    config_format_step = "Config Format Step"



import torch
from torch import nn

from .attn.attention_loss import AttentionBinarizationLoss, AttentionCTCLoss
from .config import FastSpeech2Config


class FastSpeech2Loss(nn.Module):
    def __init__(self, config: FastSpeech2Config):
        super().__init__()
        self.config = config
        self.loss_fns = {
            "mse": nn.MSELoss(),
            "mae": nn.L1Loss(),
        }
        self.attn_ctc_loss = AttentionCTCLoss()
        self.attn_bin_loss = AttentionBinarizationLoss()

    def forward(self, output, batch, current_epoch, frozen_components=None):
        # sourcery skip: merge-dict-assign, move-assign-in-block
        log_duration_prediction = output["duration_prediction"]
        duration_target = output["duration_target"]
        energy_target = output["energy_target"]
        pitch_target = output["pitch_target"]
        energy_prediction = output["energy_prediction"]
        pitch_prediction = output["pitch_prediction"]
        spec_target = batch["mel"]
        spec_prediction = output["output"]
        spec_postnet_prediction = output["postnet_output"]

        if frozen_components is None:
            frozen_components = []

        # Get masks
        src_mask = output["src_mask"]
        tgt_mask = output["tgt_mask"]

        # Don't calculate grad on target
        duration_target.requires_grad = False
        energy_target.requires_grad = False
        spec_target.requires_grad = False
        pitch_target.requires_grad = False

        losses = {}

        # Calculate pitch loss
        if self.config.model.variance_predictors.pitch.level == "phone":
            pitch_mask = src_mask
        else:
            pitch_mask = tgt_mask

        pitch_prediction = pitch_prediction * pitch_mask
        pitch_target = pitch_target * pitch_mask
        pitch_loss_fn = self.config.model.variance_predictors.pitch.loss
        losses["pitch"] = (
            self.loss_fns[pitch_loss_fn](pitch_prediction, pitch_target)
            * self.config.training.pitch_loss_weight
        )

        # Calculate energy loss
        if self.config.model.variance_predictors.energy.level == "phone":
            energy_mask = src_mask
        else:
            energy_mask = tgt_mask

        energy_prediction = energy_prediction * energy_mask
        energy_target = energy_target * energy_mask
        energy_loss_fn = self.config.model.variance_predictors.energy.loss
        losses["energy"] = (
            self.loss_fns[energy_loss_fn](energy_prediction, energy_target)
            * self.config.training.energy_loss_weight
        )

        # Calculate duration loss
        log_duration_target = torch.log(duration_target.float() + 1) * src_mask
        log_duration_prediction = log_duration_prediction * src_mask
        duration_loss_fn = self.config.model.variance_predictors.duration.loss
        losses["duration"] = (
            self.loss_fns[duration_loss_fn](
                log_duration_prediction, log_duration_target
            )
            * self.config.training.duration_loss_weight
        )

        # Calculate Mel-spectrogram loss
        tgt_mask = tgt_mask.unsqueeze(2)
        spec_prediction = spec_prediction * tgt_mask
        spec_target = spec_target * tgt_mask
        losses["spec"] = (
            self.loss_fns[self.config.model.mel_loss](spec_prediction, spec_target)
            * self.config.training.mel_loss_weight
        )
        if self.config.model.use_postnet:
            spec_postnet_prediction = spec_postnet_prediction * tgt_mask
            losses["postnet"] = (
                self.loss_fns[self.config.model.mel_loss](
                    spec_postnet_prediction, spec_target
                )
                * self.config.training.postnet_loss_weight
            )

        # Calculate attention loss if using
        if self.config.model.learn_alignment:
            ctc_loss = self.attn_ctc_loss(
                output["attn_logprob"], batch["src_lens"], batch["mel_lens"]
            )
            losses["attn_ctc"] = ctc_loss * self.config.training.attn_ctc_loss_weight
            bin_loss_weight = (
                min(
                    current_epoch / self.config.training.attn_bin_loss_warmup_epochs,
                    1.0,
                )
                * self.config.training.attn_bin_loss_weight
            )
            bin_loss = self.attn_bin_loss(output["attn_hard"], output["attn_soft"])
            losses["attn_bin"] = bin_loss * bin_loss_weight

        # Calculate total loss
        losses["total"] = sum(losses.values())
        return losses



from math import log

import torch
import torch.nn.functional as F
from torch import nn

from .blocks import ConvNorm, DepthwiseSeparableConv1d
from .utils.heavy import make_positions


class Transpose(nn.Module):
    def __init__(self, module):
        super().__init__()
        self.module = module

    def forward(self, x):
        return self.module(x.transpose(1, 2)).transpose(1, 2)


class VarianceConvolutionLayer(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: int,
        dropout: float,
        depthwise: bool,
    ):
        super().__init__()
        conv = Transpose(
            DepthwiseSeparableConv1d(in_channels, out_channels, kernel_size)
            if depthwise
            else nn.Conv1d(
                in_channels,
                out_channels,
                kernel_size,
                padding=(kernel_size - 1) // 2,
            )
        )
        self.layers = nn.Sequential(
            conv,
            nn.ReLU(),
            nn.LayerNorm(out_channels),
            nn.Dropout(dropout),
        )

    def forward(self, x):
        return self.layers(x)


class SinusoidalPositionalEmbedding(nn.Module):
    """This module produces sinusoidal positional embeddings of any length.
    Padding symbols are ignored. From: https://github.com/keonlee9420/Comprehensive-Transformer-TTS/blob/main/model/transformers/blocks.py
    """

    def __init__(self, embedding_dim, padding_idx, init_size=1024):
        super().__init__()
        self.embedding_dim = embedding_dim
        self.padding_idx = padding_idx
        self.weights = SinusoidalPositionalEmbedding.get_embedding(
            init_size,
            embedding_dim,
            padding_idx,
        )
        self.register_buffer("_float_tensor", torch.FloatTensor(1))

    @staticmethod
    def get_embedding(num_embeddings, embedding_dim, padding_idx=None):
        """Build sinusoidal embeddings.
        This matches the implementation in tensor2tensor, but differs slightly
        from the description in Section 3.5 of "Attention Is All You Need".
        """
        half_dim = embedding_dim // 2
        emb = log(10000) / (half_dim - 1)
        emb = torch.exp(torch.arange(half_dim, dtype=torch.float) * -emb)
        emb = torch.arange(num_embeddings, dtype=torch.float).unsqueeze(
            1
        ) * emb.unsqueeze(0)
        emb = torch.cat([torch.sin(emb), torch.cos(emb)], dim=1).view(
            num_embeddings, -1
        )
        if embedding_dim % 2 == 1:
            # zero pad
            emb = torch.cat([emb, torch.zeros(num_embeddings, 1)], dim=1)
        if padding_idx is not None:
            emb[padding_idx, :] = 0
        return emb

    def forward(
        self, input, incremental_state=None, timestep=None, positions=None, **kwargs
    ):
        """Input is expected to be of size [bsz x seqlen]."""
        bsz, seq_len = input.shape[:2]
        max_pos = self.padding_idx + 1 + seq_len
        if self.weights is None or max_pos > self.weights.size(0):
            # recompute/expand embeddings if needed
            self.weights = SinusoidalPositionalEmbedding.get_embedding(
                max_pos,
                self.embedding_dim,
                self.padding_idx,
            )
        self.weights = self.weights.to(self._float_tensor)

        if incremental_state is not None:
            # positions is the same for every token when decoding a single step
            pos = timestep.view(-1)[0] + 1 if timestep is not None else seq_len
            return self.weights[self.padding_idx + pos, :].expand(bsz, 1, -1)

        positions = (
            make_positions(input, self.padding_idx) if positions is None else positions
        )
        return (
            self.weights.index_select(0, positions.view(-1))
            .view(bsz, seq_len, -1)
            .detach()
        )

    def max_positions(self):
        """Maximum number of supported positions."""
        return int(1e5)  # an arbitrary large number


class PositionalEmbedding(nn.Module):
    """positional encoding from FastPitch"""

    def __init__(self, embedding_dim):
        super(PositionalEmbedding, self).__init__()
        self.demb = embedding_dim
        inv_freq = 1 / (10000 ** (torch.arange(0.0, self.demb, 2.0) / self.demb))
        self.register_buffer("inv_freq", inv_freq)

    def forward(self, pos_seq, bsz=None):
        sinusoid_inp = torch.matmul(
            torch.unsqueeze(pos_seq, -1), torch.unsqueeze(self.inv_freq, 0)
        )
        pos_emb = torch.cat([sinusoid_inp.sin(), sinusoid_inp.cos()], dim=1)
        if bsz is not None:
            return pos_emb[None, :, :].expand(bsz, -1, -1)
        else:
            return pos_emb[None, :, :]


class PostNet(nn.Module):
    """
    PostNet: Five 1-d convolution with 512 channels and kernel size 5
    """

    def __init__(
        self,
        n_mel_channels=80,
        postnet_embedding_dim=512,
        postnet_kernel_size=5,
        postnet_n_convolutions=5,
    ):
        super(PostNet, self).__init__()
        self.convolutions = nn.ModuleList()

        self.convolutions.append(
            nn.Sequential(
                ConvNorm(
                    n_mel_channels,
                    postnet_embedding_dim,
                    kernel_size=postnet_kernel_size,
                    stride=1,
                    padding=int((postnet_kernel_size - 1) / 2),
                    dilation=1,
                    w_init_gain="tanh",
                ),
                nn.BatchNorm1d(postnet_embedding_dim),
            )
        )

        for _ in range(1, postnet_n_convolutions - 1):
            self.convolutions.append(
                nn.Sequential(
                    ConvNorm(
                        postnet_embedding_dim,
                        postnet_embedding_dim,
                        kernel_size=postnet_kernel_size,
                        stride=1,
                        padding=int((postnet_kernel_size - 1) / 2),
                        dilation=1,
                        w_init_gain="tanh",
                    ),
                    nn.BatchNorm1d(postnet_embedding_dim),
                )
            )

        self.convolutions.append(
            nn.Sequential(
                ConvNorm(
                    postnet_embedding_dim,
                    n_mel_channels,
                    kernel_size=postnet_kernel_size,
                    stride=1,
                    padding=int((postnet_kernel_size - 1) / 2),
                    dilation=1,
                    w_init_gain="linear",
                ),
                nn.BatchNorm1d(n_mel_channels),
            )
        )

    def forward(self, x):
        x = x.contiguous().transpose(1, 2)

        for i in range(len(self.convolutions) - 1):
            x = F.dropout(torch.tanh(self.convolutions[i](x)), 0.5, self.training)
        x = F.dropout(self.convolutions[-1](x), 0.5, self.training)

        x = x.contiguous().transpose(1, 2)
        return x



from enum import Enum
from pathlib import Path
from typing import Annotated, Any, Dict, Optional, Union

from annotated_types import Ge
from everyvoice.config.preprocessing_config import PreprocessingConfig
from everyvoice.config.shared_types import (
    BaseModelWithContact,
    BaseTrainingConfig,
    ConfigModel,
    NoamOptimizer,
    init_context,
)
from everyvoice.config.text_config import TextConfig
from everyvoice.config.type_definitions import TargetTrainingTextRepresentationLevel
from everyvoice.config.utils import load_partials
from everyvoice.utils import load_config_from_json_or_yaml_path
from pydantic import (
    Field,
    FilePath,
    ValidationInfo,
    field_serializer,
    field_validator,
    model_validator,
)


class ConformerConfig(ConfigModel):
    layers: int = Field(4, description="The number of layers in the Conformer.")
    heads: int = Field(
        2, description="The number of heads in the multi-headed attention modules."
    )
    input_dim: int = Field(
        256,
        description="The number of hidden dimensions in the input. The input_dim value declared in the encoder and decoder modules must match the input_dim value declared in each variance predictor module.",
    )
    feedforward_dim: int = Field(
        1024, description="The number of dimensions in the feedforward layers."
    )
    conv_kernel_size: int = Field(
        9,
        description="The size of the kernel in each convoluational layer of the Conformer.",
    )
    dropout: float = Field(0.2, description="The amount of dropout to apply.")


class FastSpeech2Variances(ConfigModel):
    energy: bool = False
    duration: bool = False
    pitch: bool = False


class VarianceLevelEnum(str, Enum):
    phone = "phone"
    frame = "frame"


class VarianceLossEnum(str, Enum):
    mse = "mse"
    mae = "mae"


class VariancePredictorBase(ConfigModel):
    loss: VarianceLossEnum = Field(
        VarianceLossEnum.mse,
        description="The loss function to use when calculate variance loss. Either 'mse' or 'mae'.",
    )
    n_layers: int = Field(
        5, description="The number of layers in the variance predictor module."
    )
    kernel_size: int = Field(
        3,
        description="The kernel size of each convolutional layer in the variance predictor module.",
    )
    dropout: float = Field(0.5, description="The amount of dropout to apply.")
    input_dim: int = Field(
        256,
        description="The number of hidden dimensions in the input. This must match the input_dim value declared in the encoder and decoder modules.",
    )
    n_bins: int = Field(
        256, description="The number of bins to use in the variance predictor module."
    )
    depthwise: bool = Field(
        True, description="Whether to use depthwise separable convolutions."
    )

    @field_serializer("loss")
    def convert_loss_enum(self, loss: VarianceLossEnum):
        return loss.value


class VariancePredictorConfig(VariancePredictorBase):
    level: VarianceLevelEnum = Field(
        VarianceLevelEnum.phone,
        description="The level for the variance predictor to use. 'frame' will make predictions at the frame level. 'phone' will average predictions across all frames in each phone.",
    )

    @field_serializer("level")
    def convert_level_enum(self, level: VarianceLevelEnum):
        return level.value


class VariancePredictors(ConfigModel):
    energy: VariancePredictorConfig = Field(
        default_factory=VariancePredictorConfig,
        description="The variance predictor for energy",
    )
    duration: VariancePredictorBase = Field(
        default_factory=VariancePredictorBase,
        description="The variance predictor for duration",
    )
    pitch: VariancePredictorConfig = Field(
        default_factory=VariancePredictorConfig,
        description="The variance predictor for pitch",
    )


class FastSpeech2ModelConfig(ConfigModel):
    encoder: ConformerConfig = Field(
        default_factory=ConformerConfig,
        description="The configuration of the encoder module.",
    )
    decoder: ConformerConfig = Field(
        default_factory=ConformerConfig,
        description="The configuration of the decoder module.",
    )
    variance_predictors: VariancePredictors = Field(
        default_factory=VariancePredictors,
        description="Configuration for energy, duration, and pitch variance predictors.",
    )
    target_text_representation_level: TargetTrainingTextRepresentationLevel = (
        TargetTrainingTextRepresentationLevel.characters
    )
    learn_alignment: bool = Field(
        True,
        description="Whether to jointly learn alignments using monotonic alignment search module (See Badlani et. al. 2021: https://arxiv.org/abs/2108.10447). If set to False, you will have to provide text/audio alignments separately before training a text-to-spec (feature prediction) model.",
    )
    max_length: int = Field(
        1000, description="The maximum length (i.e. number of symbols) for text inputs."
    )
    mel_loss: VarianceLossEnum = Field(
        VarianceLossEnum.mse,
        description="The loss function to use when calculating Mel spectrogram loss.",
    )
    use_postnet: bool = Field(True, description="Whether to use a postnet module.")
    multilingual: bool = Field(
        False,
        description="Whether to train a multilingual model. For this to work, your filelist must contain a column/field for 'language' with values for each utterance.",
    )
    multispeaker: bool = Field(
        False,
        description="Whether to train a multispeaker model. For this to work, your filelist must contain a column/field for 'speaker' with values for each utterance.",
    )

    @field_serializer("mel_loss")
    def convert_mel_loss_enum(self, mel_loss: VarianceLossEnum):
        return mel_loss.value

    @field_serializer("target_text_representation_level")
    def convert_training_enum(
        self, target_text_representation_level: TargetTrainingTextRepresentationLevel
    ):
        return target_text_representation_level.value


class EarlyStoppingMetricEnum(str, Enum):
    none = "none"
    mae = "mae"
    js = "js"


class EarlyStoppingConfig(ConfigModel):
    metric: EarlyStoppingMetricEnum = EarlyStoppingMetricEnum.none
    patience: int = 4

    @field_serializer("metric")
    def convert_metric_enum(self, metric: EarlyStoppingMetricEnum):
        return metric.value


class FastSpeech2TrainingConfig(BaseTrainingConfig):
    use_weighted_sampler: bool = Field(
        False,
        description="Whether to use a sampler which oversamples from the minority language or speaker class for balanced training.",
    )
    optimizer: NoamOptimizer = Field(
        NoamOptimizer(
            betas=(0.9, 0.999), learning_rate=1e-3, warmup_steps=1000, weight_decay=1e-6
        ),
        description="The optimizer to use during training.",
    )
    # TODO: Implement early stopping
    # early_stopping: EarlyStoppingConfig = Field(default_factory=EarlyStoppingConfig)
    vocoder_path: Union[FilePath, None] = None
    mel_loss_weight: float = Field(
        1.0, description="Multiply the spec loss by this weight"
    )
    postnet_loss_weight: float = Field(
        1.0, description="Multiply the postnet loss by this weight"
    )
    pitch_loss_weight: float = Field(
        0.1, description="Multiply the pitch loss by this weight"
    )
    energy_loss_weight: float = Field(
        0.1, description="Multiply the energy loss by this weight"
    )
    duration_loss_weight: float = Field(
        0.1, description="Multiply the duration loss by this weight"
    )
    attn_ctc_loss_weight: float = Field(
        0.1, description="Multiply the Attention CTC loss by this weight"
    )
    attn_bin_loss_weight: float = Field(
        0.1, description="Multiply the Attention Binarization loss by this weight"
    )
    attn_bin_loss_warmup_epochs: Annotated[int, Ge(1)] = Field(
        100,
        description="Scale the Attention Binarization loss by (current_epoch / attn_bin_loss_warmup_epochs) until the number of epochs defined by attn_bin_loss_warmup_epochs is reached.",
    )

    # NOTE: We must use a different name for this function than the one in the
    # base class or else it overrides the base class and we lose validation.
    @field_validator("vocoder_path", mode="before")
    @classmethod
    def relative_to_absolute_vocoder_path(
        cls, value: Any, info: ValidationInfo
    ) -> Path:
        if isinstance(value, (str, Path)):
            return cls.path_relative_to_absolute(Path(value), info)
        return value


class FastSpeech2Config(BaseModelWithContact):
    model: FastSpeech2ModelConfig = Field(
        default_factory=FastSpeech2ModelConfig,
        description="The model configuration settings.",
    )
    path_to_model_config_file: Optional[FilePath] = Field(
        None, description="The path of a model configuration file."
    )

    training: FastSpeech2TrainingConfig = Field(
        default_factory=FastSpeech2TrainingConfig,
        description="The training configuration hyperparameters.",
    )
    path_to_training_config_file: Optional[FilePath] = Field(
        None, description="The path of a training configuration file."
    )

    preprocessing: PreprocessingConfig = Field(
        default_factory=PreprocessingConfig,
        description="The preprocessing configuration, including information about audio settings.",
    )
    path_to_preprocessing_config_file: Optional[FilePath] = Field(
        None, description="The path of a preprocessing configuration file."
    )

    text: TextConfig = Field(
        default_factory=TextConfig, description="The text configuration."
    )
    path_to_text_config_file: Optional[FilePath] = Field(
        None, description="The path of a text configuration file."
    )

    @model_validator(mode="before")  # type: ignore
    def load_partials(self: Dict[Any, Any], info: ValidationInfo):
        config_path = (
            info.context.get("config_path", None) if info.context is not None else None
        )
        return load_partials(
            self,
            ("model", "training", "preprocessing", "text"),
            config_path=config_path,
        )

    @staticmethod
    def load_config_from_path(path: Path) -> "FastSpeech2Config":
        """Load a config from a path"""
        config = load_config_from_json_or_yaml_path(path)
        with init_context({"config_path": path}):
            config = FastSpeech2Config(**config)
        return config

    # INPUT_TODO: initialize text with union of symbols from dataset



import math
import random
from typing import Tuple

import numpy as np
import torch
from everyvoice.utils.heavy import get_spectral_transform
from loguru import logger

from .config import HiFiGANConfig
from .model import HiFiGAN


def load_hifigan_from_checkpoint(ckpt: dict, device) -> Tuple[HiFiGAN, HiFiGANConfig]:
    config: dict | HiFiGANConfig = ckpt["hyper_parameters"]["config"]
    if isinstance(config, dict):
        config = HiFiGANConfig(**config)
    model = HiFiGAN(config).to(device)
    model.load_state_dict(ckpt["state_dict"])
    model.generator.eval()
    model.generator.remove_weight_norm()
    return model, config


def synthesize_data(
    data: torch.Tensor, model: HiFiGAN, config: HiFiGANConfig
) -> Tuple[np.ndarray, int]:
    """Synthesize a batch of waveforms from spectral features

    Args:
        data (Tensor): data tensor, expects output from feature prediction network to be size (b=batch_size, t=number_of_frames, k=n_mels)
        ckpt (dict): HiFiGANLightning checkpoint, expects checkpoint to have a 'hyper_parameters.config' key and HiFiGANConfig object value as well as a 'state_dict' key with model weight as the value
    Returns:
        Tuple[np.ndarray, int]: a B, T array of the synthesized audio and the sampling rate
    """
    if config.model.istft_layer:
        inverse_spectral_transform = get_spectral_transform(
            "istft",
            model.generator.post_n_fft,
            model.generator.post_n_fft,
            model.generator.post_n_fft // 4,
        ).to(data.device)
        with torch.no_grad():
            mag, phase = model.generator(data.transpose(1, 2))
        # We can remove this once the fix for https://github.com/pytorch/pytorch/issues/119088 is merged
        if mag.device.type == "mps" or phase.device.type == "mps":
            logger.warning(
                "Handling complex numbers is broken on MPS (last checked in torch==2.2.0), so we are falling back to CPU"
            )
            mag = mag.to("cpu")
            phase = phase.to("cpu")
            inverse_spectral_transform.to("cpu")
        wavs = inverse_spectral_transform(mag * torch.exp(phase * 1j)).unsqueeze(-2)
    else:
        with torch.no_grad():
            wavs = model.generator(data.transpose(1, 2))
    # squeeze to remove the channel dimension
    return (
        wavs.squeeze(1).cpu().numpy(),
        config.preprocessing.audio.output_sampling_rate,
    )


def get_all_segments(
    x: torch.Tensor, y: torch.Tensor, y_mel: torch.Tensor, segment_size, output_hop_size
):
    """Randomly select a segment from y (wav), x (spectrogram), and y_mel (spectrogram), if the segment is too short, pad it with zeros

    Args:
        x (torch.Tensor): spectrogram input of vocoder
        y (torch.Tensor): waveform output of vocoder
        y_mel (torch.Tensor): spectrogram of waveform output of vocoder
    """
    # segment size is relative to output_sampling_rate, so we use the output_hop_size, but frames_per_seg is in frequency domain, so invariant to x and y_mel
    # other implementations just resample y and take the mel spectrogram of that, but this solution allows for segmenting predicted mel spectrograms from the acoustic feature prediction network too
    frames_per_seg = math.ceil(segment_size / output_hop_size)
    if y.size(0) > segment_size:
        max_spec_start = x.size(1) - frames_per_seg - 1
        spec_start = random.randint(0, max_spec_start)
        x = x[:, spec_start : spec_start + frames_per_seg]
        y_mel = y_mel[:, spec_start : spec_start + frames_per_seg]
        y = y[
            spec_start
            * output_hop_size : (spec_start + frames_per_seg)
            * output_hop_size,
        ]
    else:
        x = torch.nn.functional.pad(x, (0, frames_per_seg - x.size(1)), "constant")
        y_mel = torch.nn.functional.pad(
            y_mel,
            (0, frames_per_seg - y_mel.size(1)),
            "constant",
        )
        y = torch.nn.functional.pad(y, (0, segment_size - y.size(0)), "constant")
    return x, y, y_mel



"""
Types shared among all EV modules, though not always super fast to load.
Tiny type defs that load in milliseconds can go into type_definitions.py.
"""

from collections.abc import Mapping, Sequence
from contextlib import contextmanager
from contextvars import ContextVar
from functools import cached_property
from pathlib import Path
from typing import Any, Dict, Iterator, Tuple, Union

from pydantic import (
    BaseModel,
    ConfigDict,
    EmailStr,
    Field,
    ValidationInfo,
    model_validator,
)
from typing_extensions import Annotated, Optional

from everyvoice.config.utils import (
    PossiblyRelativePath,
    PossiblyRelativePathMustExist,
    PossiblySerializedCallable,
)
from everyvoice.utils import generic_psv_filelist_reader, get_current_time

_init_context_var = ContextVar("_init_context_var", default=None)


@contextmanager
def init_context(value: Dict[str, Any]) -> Iterator[None]:
    """
    This context manager is use in tandem with pydantic to pass down a context to validators.
    As described in [Using validation context with BaseModel initialization]
    (https://docs.pydantic.dev/2.0/usage/validators/#using-validation-context-with-basemodel-initialization):
    Although there is no way to specify a context in the standard BaseModel
    initializer, you can work around this through the use of
    contextvars.ContextVar and a custom __init__ method.
    """
    token = _init_context_var.set(value)  # type: ignore
    try:
        yield
    finally:
        _init_context_var.reset(token)


class ConfigModel(BaseModel):
    model_config = ConfigDict(
        extra="forbid",
        json_schema_extra={"$schema": "http://json-schema.org/draft-07/schema#"},
    )

    def model_checkpoint_dump(self):
        """When saving a configuration to a checkpoint, we have to ensure that it is:

            - A JSON-serializable dict, to allow interoperability between different
              versions of EveryVoice and Pydantic.
            - Contains no Paths, which are validated by Pydantic and throw ValidationErrors
              if the path does not exist. Checkpoints by their nature will be shared
              across different environments so we cannot guarantee the presence of any Path
              used during initial training.

        Therefore we dump the model checkpoint (which will serialize functions into a string format)
        and then recursively search over the dumped checkpoint to remove fields with Path values.

        Returns:
            dict: A JSON-serializable dict containing no paths
        """
        ckpt = (
            self.model_dump()
        )  # can't use mode='json' because we need to filter Paths first

        def delete_paths(ckpt: dict | list | tuple):
            if isinstance(ckpt, dict):
                for k, v in ckpt.copy().items():
                    if not isinstance(v, Path):
                        ckpt[k] = delete_paths(v)
                    else:
                        del ckpt[k]
            if isinstance(ckpt, list):
                return [delete_paths(x) for x in ckpt if not isinstance(x, Path)]
            if isinstance(ckpt, tuple):
                return tuple(delete_paths(x) for x in ckpt if not isinstance(x, Path))
            else:
                return ckpt

        return delete_paths(ckpt)

    def update_config(self, new_config: dict):
        """Update the config with new values"""
        new_data = self.combine_configs(dict(self), new_config)
        self.__init__(**new_data)  # type: ignore
        return self

    @staticmethod
    def combine_configs(orig_dict: Union[dict, Sequence], new_dict: Mapping):
        """See https://stackoverflow.com/questions/3232943/update-value-of-a-nested-dictionary-of-varying-depth"""
        if isinstance(orig_dict, Sequence):
            orig_list = list(orig_dict)
            for key_s, val in new_dict.items():
                key_i = int(key_s)
                if isinstance(val, Mapping):
                    tmp = ConfigModel.combine_configs(orig_list[key_i], val)
                    orig_list[key_i] = tmp
                else:
                    orig_list[key_i] = val
            return orig_list

        orig_dict = dict(orig_dict)
        new_dict = dict(new_dict)
        for key, val in new_dict.items():
            if isinstance(val, Mapping):
                tmp = ConfigModel.combine_configs(orig_dict.get(key, {}), val)
                orig_dict[key] = tmp
            else:
                orig_dict[key] = new_dict[key]
        return orig_dict


class PartialLoadConfig(ConfigModel):
    """Models that have partial models which requires a context to properly load."""

    # [Using validation context with BaseModel initialization](https://docs.pydantic.dev/2.3/usage/validators/#using-validation-context-with-basemodel-initialization)
    def __init__(__pydantic_self__, **data: Any) -> None:
        __pydantic_self__.__pydantic_validator__.validate_python(
            data,
            self_instance=__pydantic_self__,
            context=_init_context_var.get(),
        )

    @classmethod
    def path_relative_to_absolute(cls, value: Path | str, info: ValidationInfo) -> Path:
        """
        Given `value` a relative path from `config` and a `config_path` defined in `info`, transform `value` into an absolute path.
        """
        # Make sure value is a path because it can be a string when we load a model that is not partial.
        path = Path(value)
        if info.context and path is not None and not path.is_absolute():
            config_path = info.context.get("config_path", Path("."))
            path = (config_path.parent / path).resolve()
        return path


class LoggerConfig(PartialLoadConfig):
    """The logger configures all the information needed for where to store your experiment's logs and checkpoints.
    The structure of your logs will then be:
    <name> / <version> / <sub_dir>
    <sub_dir> will be generated by calling <sub_dir_callable> each time the LoggerConfig is constructed.
    """

    name: str = Field(
        "BaseExperiment",
        title="Experiment Name",
        description="The name of the experiment. The structure of your logs will be <name> / <version> / <sub_dir>.",
    )

    save_dir: PossiblyRelativePathMustExist = Field(
        Path("./logs_and_checkpoints"),
        description="The directory to save your checkpoints and logs to.",
    )

    sub_dir_callable: PossiblySerializedCallable = Field(
        get_current_time,
        description="The function that generates a string to call your runs - by default this is a timestamp. The structure of your logs will be <name> / <version> / <sub_dir> where <sub_dir> is a timestamp.",
    )

    version: str = Field(
        "base",
        description="The version of your experiment. The structure of your logs will be <name> / <version> / <sub_dir>.",
    )

    @cached_property
    def sub_dir(self) -> str:
        return self.sub_dir_callable()


class BaseTrainingConfig(PartialLoadConfig):
    batch_size: int = Field(
        16,
        description="The number of samples to include in each batch when training. If you are running out of memory, consider lowering your batch_size.",
    )
    save_top_k_ckpts: int = Field(5, description="The number of checkpoints to save.")
    # According to
    # [ModelCheckpoint](https://lightning.ai/docs/pytorch/stable/api/lightning.pytorch.callbacks.ModelCheckpoint.html#lightning.pytorch.callbacks.ModelCheckpoint),
    # ckpt_epochs and ckpt_steps must be None or non-negative.
    # 0 is the same as None, and disables checkpointing
    ckpt_steps: Union[Annotated[int, Field(ge=0)], None] = Field(
        None,
        description="The interval (in steps) for saving a checkpoint. By default checkpoints are saved every epoch using the 'ckpt_epochs' hyperparameter",
    )
    ckpt_epochs: Union[Annotated[int, Field(ge=0)], None] = Field(
        1,
        description="The interval (in epochs) for saving a checkpoint. You can also save checkpoints after n steps by using 'ckpt_steps'",
    )
    val_check_interval: Union[int, float, None] = Field(
        500,
        description="How often to check the validation set."
        " Pass a float in the range [0.0, 1.0] to check after a fraction of the training epoch."
        " Pass an int to check after a fixed number of training batches.",
    )
    check_val_every_n_epoch: Optional[int] = Field(
        None,
        description="Run validation after every n epochs. Defaults to 1, but if you have a small dataset you should change this to be larger to speed up training",
    )
    max_epochs: int = Field(1000, description="Stop training after this many epochs")
    max_steps: int = Field(100000, description="Stop training after this many steps")
    finetune_checkpoint: Union[PossiblyRelativePath, None] = Field(
        None,
        description="Automatically resume training from a checkpoint loaded from this path.",
    )
    training_filelist: PossiblyRelativePath = Field(
        Path("./path/to/your/preprocessed/training_filelist.psv"),
        description="The path to a filelist containing samples belonging to your training set.",
    )
    validation_filelist: PossiblyRelativePath = Field(
        Path("./path/to/your/preprocessed/validation_filelist.psv"),
        description="The path to a filelist containing samples belonging to your validation set.",
    )
    filelist_loader: PossiblySerializedCallable = Field(
        generic_psv_filelist_reader,
        description="Advanced. The function to use to load the filelist.",
    )
    logger: LoggerConfig = Field(
        default_factory=LoggerConfig,
        description="The configuration for the logger.",
    )
    val_data_workers: int = Field(
        0,
        description="The number of CPU workers to use when loading data during validation.",
    )
    train_data_workers: int = Field(
        4,
        description="The number of CPU workers to use when loading data during training.",
    )

    @model_validator(mode="after")
    def multually_exclusive_ckpt_options(self) -> "BaseTrainingConfig":
        """
        As documented in
        [ModelCheckpoint](https://lightning.ai/docs/pytorch/stable/api/lightning.pytorch.callbacks.ModelCheckpoint.html#lightning.pytorch.callbacks.ModelCheckpoint),
        `ckpt_steps` and `ckpt_epochs` have to be mutually exclusive.
        """
        if self.ckpt_epochs is not None and self.ckpt_steps is not None:
            raise ValueError("ckpt_epochs and ckpt_steps have to be mutually exclusive")
        return self


class ContactInformation(ConfigModel):
    contact_name: str = Field(
        description="The name of the contact person or organization responsible for answering questions related to this model."
    )
    contact_email: EmailStr = Field(
        description="The email address of the contact person or organization responsible for answering questions related to this model."
    )


class BaseModelWithContact(PartialLoadConfig):
    # TODO: finish guide in docs
    contact: ContactInformation = Field(
        description="EveryVoice requires a contact name and email to help prevent misuse. Please read our Guide <https://docs.everyvoice.ca/latest/> to understand more about the importance of misuse prevention with TTS."
    )


class BaseOptimizer(ConfigModel):
    learning_rate: float = Field(1e-4, description="The initial learning rate to use")
    eps: float = Field(
        1e-8,
        description="Advanced. The value of optimizer constant Epsilon, used for numerical stability.",
    )
    weight_decay: float = 0.01


class RMSOptimizer(BaseOptimizer):
    alpha: float = Field(
        0.99,
        description="Advanced. The value of RMSProp optimizer alpha smoothing constant.",
    )
    name: str = Field("rms", description="The name of the optimizer to use.")


class AdamOptimizer(BaseOptimizer):
    betas: Tuple[float, float] = Field(
        (0.9, 0.98),
        description="Advanced. The values of the Adam Optimizer beta coefficients.",
    )
    name: str = Field("adam", description="The name of the optimizer to use.")


class AdamWOptimizer(BaseOptimizer):
    betas: Tuple[float, float] = Field(
        (0.9, 0.98),
        description="Advanced. The values of the AdamW Optimizer beta coefficients.",
    )
    name: str = Field("adamw", description="The name of the optimizer to use.")


class NoamOptimizer(AdamOptimizer):
    warmup_steps: int = Field(
        1000,
        description="The number of steps to increase the learning rate before starting to decrease it.",
    )
    name: str = Field("noam", description="The name of the optimizer to use.")



#!/usr/bin/env python

""" Organize tests into Test Suites
"""

import importlib
import os
import re
import sys
from typing import Iterable
from unittest import TestLoader, TestSuite, TextTestRunner

from loguru import logger

# Unit tests

SUBMODULE_SUITES: dict[str, tuple[str, ...]] = {
    "fs2": ("/model/feature_prediction/FastSpeech2_lightning/fs2/tests",),
}
SUITES: dict[str, tuple[str, ...]] = {
    "config": ("test_configs",),
    "loader": ("test_dataloader",),
    "text": ("test_text", "test_utils"),
    "preprocessing": ("test_preprocessing",),
    "model": ("test_model",),
    "cli": ("test_wizard", "test_cli"),
    **SUBMODULE_SUITES,
}
dev_suites = ("config", "loader", "text", "preprocessing", "model", "cli", "fs2")
SUITES["dev"] = sum((SUITES[suite] for suite in dev_suites), start=())


def remove_test_prefix(test_case: str):
    for prefix in "<", "everyvoice.", "tests.":
        if test_case.startswith(prefix):
            test_case = test_case[len(prefix) :]
    return "<" + test_case


def list_tests(suite: TestSuite):
    for subsuite in suite:
        # print(str(subsuite))
        for match in re.finditer(r"tests=\[([^][]+)\]>", str(subsuite)):
            for test_case in match[1].split(", "):
                yield remove_test_prefix(test_case)


def all_test_suites() -> TestSuite:
    loader = TestLoader()
    # NOTE: Looking specifically under `/tests` removes empty TestSuites.
    test_suite = loader.discover(
        os.path.dirname(__file__) + "/tests",
        top_level_dir=os.path.dirname(__file__),
    )
    for submodule_testsuite in SUBMODULE_SUITES.values():
        suite = loader.discover(
            os.path.dirname(__file__) + submodule_testsuite[0],
        )
        test_suite.addTests(suite)

    return test_suite


def describe_suite(suite: TestSuite):
    full_suite = all_test_suites()
    full_list = list(list_tests(full_suite))
    requested_list = list(list_tests(suite))
    requested_set = set(requested_list)
    print("Test suite includes:", *sorted(requested_list), sep="\n")
    print(
        "\nTest suite excludes:",
        *sorted(test for test in full_list if test not in requested_set),
        sep="\n",
    )


def run_tests(suite: str, describe: bool = False):
    """Decide which Test Suite to run"""
    logger.info(f"Loading test suite '{suite}'. This may take a while...")
    if suite == "all":
        test_suite = all_test_suites()
    else:
        loader = TestLoader()
        tests: Iterable[str]
        if suite in SUITES:
            tests = SUITES[suite]
        else:
            logger.error(
                f"Please specify a test suite to run: one of '{['all'] + sorted(SUITES.keys())}'."
            )
            return False
        tests = [
            "everyvoice.tests." + test if not test.startswith("/") else test
            for test in tests
        ]
        test_suite = TestSuite()
        for test in tests:
            logger.info(f"Loading {test=}")
            if test.startswith("/"):
                sub_suite = loader.discover(
                    os.path.dirname(__file__) + test,
                    top_level_dir=os.path.dirname(__file__),  # MANDATORY
                )
                test_suite.addTests(sub_suite)
            else:
                importlib.import_module(test)
                test_suite.addTest(loader.loadTestsFromName(test))

    if describe:
        describe_suite(test_suite)
        return True
    else:
        logger.info("Running test suite")
        return TextTestRunner(verbosity=3).run(test_suite).wasSuccessful()


if __name__ == "__main__":
    describe = "--describe" in sys.argv
    if describe:
        sys.argv.remove("--describe")

    try:
        suite = sys.argv[1]
    except IndexError:
        logger.info('No test suite specified, defaulting to "dev"')
        suite = "dev"
    result = run_tests(suite, describe)
    if not result:
        sys.exit(1)



import io
import logging
from contextlib import contextmanager, redirect_stderr, redirect_stdout
from typing import Any, Generator, Sequence, Union

from loguru import logger

from everyvoice.wizard import prompts

_NOTSET = object()  # Sentinel object for monkeypatch()


@contextmanager
def monkeypatch(obj, name, value) -> Generator:
    """Monkey patch obj.name to value for the duration of the context manager's life.

    Yields:
        value: the value monkey-patched, for use with "as v" notation"""
    saved_value = getattr(obj, name, _NOTSET)
    try:
        setattr(obj, name, value)
        yield value
    finally:
        if saved_value is _NOTSET:  # pragma: no cover
            delattr(obj, name)
        else:
            setattr(obj, name, saved_value)


QUIET = logging.CRITICAL + 100  # level high enough to disable all logging


@contextmanager
def patch_logger(
    module, level: int = logging.INFO
) -> Generator[logging.Logger, None, None]:
    """Monkey patch the logger for a given module with a unit testing logger
    of the given level. Use level=QUIET to just silence logging.

    Yields:
        logger (logging.Logger): patched logger, e.g., for use in self.assertLogs(logger)
    """
    with monkeypatch(module, "logger", logging.getLogger("UnitTesting")) as logger:
        logger.setLevel(level)
        yield logger


@contextmanager
def mute_logger(module: str) -> Generator[None, None, None]:
    """
    Temporarily mutes a module's `logger`.

    with mute_logger("everyvoice.base_cli.helpers"):
        config = FastSpeech2Config()
    """
    logger.disable(module)
    yield
    logger.enable(module)


@contextmanager
def capture_logs():
    """
    Context manager to capture log messages from loguru.
    """
    # [How to test loguru logger with unittest?](https://github.com/Delgan/loguru/issues/616)
    output = []
    handler_id = logger.add(output.append)
    yield output
    logger.remove(handler_id)


@contextmanager
def capture_stdout() -> Generator[io.StringIO, None, None]:
    """Context manager to capture what is printed to stdout.

    Usage:
        with capture_stdout() as stdout:
            # do stuff whose stdout you want to capture
        stdout.getvalue() is what was printed to stdout during the context

        with capture_stdout():
            # do stuff with stdout suppressed

    Yields:
        stdout (io.StringIO): captured stdout
    """
    f = io.StringIO()
    with redirect_stdout(f):
        yield f


@contextmanager
def capture_stderr():
    """Context manager to capture what is printed to stderr.

    Usage:
        with capture_stderr() as stderr:
            # do stuff whose stderr you want to capture
        stderr.getvalue() is what was printed to stderr during the context

        with capture_stderr():
            # do stuff with stderr suppressed

    Yields:
        stderr (io.StringIO): captured stderr
    """
    f = io.StringIO()
    with redirect_stderr(f):
        yield f


@contextmanager
def patch_menu_prompt(
    response_index: Union[int, list],
    multi=False,
) -> Generator[io.StringIO, None, None]:
    """Context manager to simulate what option(s) the user selects in a simple_term_menu.

    Args:
        response_index: the user's choice as a zero-based index for a
            single-option menu, or the list of indices for a multi-option
            menu
        multi: if True, response_index is a list of response_indices used one
            after the other each time a new simple_term_menu is instantiated

    Yields:
        stdout (io.StringIO): captured stdout stream.
    """
    with capture_stdout() as stdout:
        with monkeypatch(
            prompts, "simple_term_menu", SimpleTermMenuStub(response_index, multi)
        ):
            yield stdout


@contextmanager
def null_patch() -> Generator[None, None, None]:
    """dummy context manager when we must pass a monkeypatch but have nothing to patch"""
    yield


class Say:
    """Mock callable that returns response (if multi=False) or each value in
    response in turn (if multi=True) when it is called."""

    def __init__(self, response, multi=False) -> None:
        self.response = response
        self.last_index = -1
        self.multi = multi

    def __call__(self, *args: Any, **kwds: Any) -> Any:
        if self.multi:
            self.last_index += 1
            response = self.response[self.last_index]
        else:
            response = self.response
        if isinstance(response, BaseException):
            raise response
        return response


class SimpleTermMenuStub:
    """Stub class for the simple_term_menu module."""

    def __init__(self, response: Union[int, list[int]], multi=False):
        """Constructor

        Args:
            response: the index or indices the user will be simulated to choose
        """
        self.multi = multi
        self.last_index = -1
        self.response = response

    def TerminalMenu(self, *args, **kwargs):
        if not kwargs.get("raise_error_on_interrupt", False):  # pragma: no cover
            raise Exception(
                "raise_error_on_interrupt=True is required for TerminalMenu so we can receive and handle KeyboardInterrupt correctly"
            )
        return self

    def show(self):
        if self.multi:
            print("term stub", self.last_index, self.response[self.last_index + 1])
            self.last_index += 1
            response = self.response[self.last_index]
        else:
            response = self.response
        if isinstance(response, BaseException):
            raise response
        return response


class QuestionaryStub:
    """Stub class for the questionary module"""

    def __init__(self, responses: Sequence[str]) -> None:
        """Constructor

        Args:
            responses (str): the sequence of answers the user is simulated to provide
        """
        self.last_index = -1
        self.responses = responses

    def path(self, *args, **kwargs):
        return self

    text = path

    def ask(self):  # pragma: no cover
        # This will trigger a unit test failure if we use .ask()
        raise Exception(
            "Always use unsafe_ask() for questionary instances so that KeyboardInterrupt gets passed up to us."
        )

    def unsafe_ask(self):
        self.last_index += 1
        response = self.responses[self.last_index]
        if isinstance(response, BaseException):
            raise response
        return response



# Copyright (c) 2021, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Source: https://arxiv.org/abs/2108.10447

import numpy as np
import torch
from torch import nn
from torch.nn import functional as F


class ConvNorm(torch.nn.Module):
    def __init__(
        self,
        in_channels,
        out_channels,
        kernel_size=1,
        stride=1,
        padding=None,
        dilation=1,
        bias=True,
        w_init_gain="linear",
    ):
        super(ConvNorm, self).__init__()
        if padding is None:
            assert kernel_size % 2 == 1
            padding = int(dilation * (kernel_size - 1) / 2)

        self.conv = torch.nn.Conv1d(
            in_channels,
            out_channels,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
            dilation=dilation,
            bias=bias,
        )

        torch.nn.init.xavier_uniform_(
            self.conv.weight, gain=torch.nn.init.calculate_gain(w_init_gain)
        )

    def forward(self, signal):
        conv_signal = self.conv(signal)
        return conv_signal


class Invertible1x1ConvLUS(torch.nn.Module):
    def __init__(self, c):
        super(Invertible1x1ConvLUS, self).__init__()
        # Sample a random orthonormal matrix to initialize weights
        W, _ = torch.linalg.qr(torch.randn(c, c))
        # Ensure determinant is 1.0 not -1.0
        if torch.det(W) < 0:
            W[:, 0] = -1 * W[:, 0]
        # Original implementation is deprecated: https://pytorch.org/docs/stable/generated/torch.lu.html
        # p, lower, upper = torch.lu_unpack(*torch.linalg.lu_factor(W))
        p, lower, upper = torch.linalg.lu(W)

        self.register_buffer("p", p)
        # diagonals of lower will always be 1s anyway
        lower = torch.tril(lower, -1)
        lower_diag = torch.diag(torch.eye(c, c))
        self.register_buffer("lower_diag", lower_diag)
        self.lower = nn.Parameter(lower)
        self.upper_diag = nn.Parameter(torch.diag(upper))
        self.upper = nn.Parameter(torch.triu(upper, 1))

    def forward(self, z, reverse=False):
        U = torch.triu(self.upper, 1) + torch.diag(self.upper_diag)
        L = torch.tril(self.lower, -1) + torch.diag(self.lower_diag)
        W = torch.mm(self.p, torch.mm(L, U))
        if reverse:
            if not hasattr(self, "W_inverse"):
                # Reverse computation
                W_inverse = W.float().inverse()
                if z.type() == "torch.cuda.HalfTensor":
                    W_inverse = W_inverse.half()

                self.W_inverse = W_inverse[..., None]
            z = F.conv1d(z, self.W_inverse, bias=None, stride=1, padding=0)
            return z
        else:
            W = W[..., None]
            z = F.conv1d(z, W, bias=None, stride=1, padding=0)
            log_det_W = torch.sum(torch.log(torch.abs(self.upper_diag)))
            return z, log_det_W


class ConvAttention(torch.nn.Module):
    def __init__(
        self,
        n_mel_channels=80,
        n_speaker_dim=128,
        n_text_channels=512,
        n_att_channels=80,
        temperature=1.0,
        n_mel_convs=2,
        align_query_enc_type="3xconv",
        use_query_proj=True,
    ):
        super(ConvAttention, self).__init__()
        self.temperature = temperature
        self.att_scaling_factor = np.sqrt(n_att_channels)
        self.softmax = torch.nn.Softmax(dim=3)
        self.log_softmax = torch.nn.LogSoftmax(dim=3)
        self.query_proj = Invertible1x1ConvLUS(n_mel_channels)
        self.align_query_enc_type = align_query_enc_type
        self.use_query_proj = bool(use_query_proj)

        self.key_proj = nn.Sequential(
            ConvNorm(
                n_text_channels,
                n_text_channels * 2,
                kernel_size=3,
                bias=True,
                w_init_gain="relu",
            ),
            torch.nn.ReLU(),
            ConvNorm(n_text_channels * 2, n_att_channels, kernel_size=1, bias=True),
        )

        self.align_query_enc_type = align_query_enc_type

        if align_query_enc_type == "inv_conv":
            self.query_proj = Invertible1x1ConvLUS(n_mel_channels)
        elif align_query_enc_type == "3xconv":
            self.query_proj = nn.Sequential(
                ConvNorm(
                    n_mel_channels,
                    n_mel_channels * 2,
                    kernel_size=3,
                    bias=True,
                    w_init_gain="relu",
                ),
                torch.nn.ReLU(),
                ConvNorm(n_mel_channels * 2, n_mel_channels, kernel_size=1, bias=True),
                torch.nn.ReLU(),
                ConvNorm(n_mel_channels, n_att_channels, kernel_size=1, bias=True),
            )
        else:
            raise ValueError("Unknown query encoder type specified")

    def run_padded_sequence(
        self, sorted_idx, unsort_idx, lens, padded_data, recurrent_model
    ):
        """Sorts input data by previded ordering (and un-ordering) and runs the
        packed data through the recurrent model

        Args:
            sorted_idx (torch.tensor): 1D sorting index
            unsort_idx (torch.tensor): 1D unsorting index (inverse of sorted_idx)
            lens: lengths of input data (sorted in descending order)
            padded_data (torch.tensor): input sequences (padded)
            recurrent_model (nn.Module): recurrent model to run data through
        Returns:
            hidden_vectors (torch.tensor): outputs of the RNN, in the original,
            unsorted, ordering
        """

        # sort the data by decreasing length using provided index
        # we assume batch index is in dim=1
        padded_data = padded_data[:, sorted_idx]
        padded_data = nn.utils.rnn.pack_padded_sequence(padded_data, lens)
        hidden_vectors = recurrent_model(padded_data)[0]
        hidden_vectors, _ = nn.utils.rnn.pad_packed_sequence(hidden_vectors)
        # unsort the results at dim=1 and return
        hidden_vectors = hidden_vectors[:, unsort_idx]
        return hidden_vectors

    def encode_query(self, query, query_lens):
        query = query.permute(2, 0, 1)  # seq_len, batch, feature dim
        lens, ids = torch.sort(query_lens, descending=True)
        original_ids = [0] * lens.size(0)
        for i in range(len(ids)):
            original_ids[ids[i]] = i

        query_encoded = self.run_padded_sequence(
            ids, original_ids, lens, query, self.query_lstm
        )
        query_encoded = query_encoded.permute(1, 2, 0)
        return query_encoded

    def forward(
        self,
        queries,
        keys,
        query_lens,
        mask=None,
        key_lens=None,
        keys_encoded=None,
        attn_prior=None,
    ):
        """Attention mechanism for flowtron parallel
        Unlike in Flowtron, we have no restrictions such as causality etc,
        since we only need this during training.

        Args:
            queries (torch.tensor): B x C x T1 tensor
                (probably going to be mel data)
            keys (torch.tensor): B x C2 x T2 tensor (text data)
            query_lens: lengths for sorting the queries in descending order
            mask (torch.tensor): uint8 binary mask for variable length entries
                (should be in the T2 domain)
        Output:
            attn (torch.tensor): B x 1 x T1 x T2 attention mask.
                Final dim T2 should sum to 1
        """
        keys_enc = self.key_proj(keys)  # B x n_attn_dims x T2

        # Beware can only do this since query_dim = attn_dim = n_mel_channels
        if self.use_query_proj:
            if self.align_query_enc_type == "inv_conv":
                queries_enc, _ = self.query_proj(queries)
            elif self.align_query_enc_type == "3xconv":
                queries_enc = self.query_proj(queries)
                # log_det_W = 0.0
            else:
                queries_enc, _ = self.query_proj(queries)
        else:
            queries_enc, _ = queries, 0.0

        # different ways of computing attn,
        # one is isotopic gaussians (per phoneme)
        # Simplistic Gaussian Isotopic Attention

        # B x n_attn_dims x T1 x T2
        attn = (queries_enc[:, :, :, None] - keys_enc[:, :, None]) ** 2
        # compute log likelihood from a gaussian
        attn = -0.0005 * attn.sum(1, keepdim=True)
        if attn_prior is not None:
            attn = self.log_softmax(attn) + torch.log(attn_prior[:, None] + 1e-8)

        attn_logprob = attn.clone()

        if mask is not None:
            attn.data.masked_fill_(mask.permute(0, 2, 1).unsqueeze(2), -float("inf"))

        attn = self.softmax(attn)  # Softmax along T2
        return attn, attn_logprob
